# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
This is 'n_oct' flattened octahedron net xy
"""

import numpy as np
from numpy.typing import NDArray
from hhg9.base.composite import CompositeDomain, ComponentDomain
from hhg9.base.point_format import PointFormat
from hhg9.projections import BaryNet
from hhg9.domains.nets import net_layouts
from hhg9.h9 import H9K, H9P
from hhg9.algorithms.geometry import inside_triangle_cw


class OctantNet(ComponentDomain):
    """
    This a 2D side of an Octant that belongs to a Net.
    Validity should be easy enough since we have the 3 points that define it.
    """
    def __init__(self, registrar, dom, name: str, sign: tuple, mode: int):
        super().__init__(registrar, name, dom, mode, sign, 2)

    def valid(self, pts: NDArray) -> NDArray:
        """
        Return an array of bools according to the validity criterion
        :param pts: set of 3d Euclidean points
        """
        raise NotImplementedError


class OctahedralNet(CompositeDomain):
    """
    This is a 2d net correlate of the Octahedron.
    Triangles have edge length √2 in a unit octahedron
    """
    from hhg9 import Points

    R3 = H9K.R3
    GW = H9K.lattice.U * 3  # grid unit width U = H9GC.U H9GC.W/6
    GH = H9K.lattice.V * 3  # grid unit height
    RT = np.pi / 3.      # grid rotation in 60º

    def __init__(self, registrar, *, layout='mortar'):
        c_oct = registrar.domain('c_oct')
        b_oct = registrar.domain('b_oct')
        if layout not in net_layouts:
            layout = 'mortar'
        super().__init__(registrar, f'n_oct:{layout}', 2)
        tp = H9P.sv  # mode vertices
        self.sides = {}
        self.projs = {}
        self.face_tris = {}
        self.sign_to_side = {}
        self.c_oct = c_oct
        self.b_oct = b_oct
        self.layout = net_layouts[layout]
        grid_xy = np.array(list(self.layout['grid'].values()))[:, :2]
        # gx_min, gx_max = grid_xy[:, 0].min(), grid_xy[:, 0].max()
        # gy_min, gy_max = grid_xy[:, 1].min(), grid_xy[:, 1].max()

        # Each placed face contributes [x_off, x_off + 2*GW] and [y_off, y_off + 3*GH]
        # So total width/height is span of gx,gy plus the single-face width/height
        # self.wi = (gx_max - gx_min + 2) * self.GW
        # self.he = (gy_max - gy_min + 3) * self.GH
        self.wi = self.layout['width'] * H9K.derived.W
        self.he = self.layout['height'] * H9K.derived.H

        self.oid_mo = np.zeros((8,), dtype=np.uint8)
        for sign, val in self.layout['grid'].items():
            side = self.c_oct.signs[sign]
            oid = self.c_oct.face_id[side]
            bary = b_oct.sides[side]
            gx, gy, th = val
            x_off = gx * self.GW
            y_off = gy * self.GH
            n_theta = (th % 6) * self.RT
            flipped = th % 2
            mode = {0: 1, 1: 0}[bary.mode] if flipped else bary.mode
            n_sig = f'{self.name}:{side}'
            b_sig = f'{b_oct.name}:{side}'
            self.oid_mo[oid] = mode
            self.sides[sign] = OctantNet(registrar, self, n_sig, sign, mode)
            face = BaryNet(registrar, side, b_sig, n_sig, n_theta, (x_off, y_off))
            self.projs[side] = face
            tri = H9P.sv[bary.mode]  # triangle from H9P. Use bary.mo b/c will transform!
            tri_rt = tri @ face.matrix + face.offset  # bary->net
            # Map sign→triangle and sign→side for fast lookup
            self.face_tris[sign] = tri_rt
            self.sign_to_side[sign] = side
            if 'c2' in self.layout:
                c2f = self.layout['c2'][sign]
                c2x = [(x * self.GW, y * self.GH, (t % 6) * self.RT) for (x, y, t) in c2f]
                face.c2trans = c2x
            self.components[sign] = self.sides[sign]


    def ratio(self):
        """Return width/height ratio"""
        return self.wi/self.he

    def img_adj(self):
        """
        :return: w,h adjustment to pixels (subtracted when outputting to image)
        """
        return self.layout['width'] + 0.51, self.layout['height'] + 0.51

    def image_dims(self, pixels: int):
        """Given the side of a triangle in pixels, return the image dimensions."""
        tri_w = pixels
        tri_h = pixels * self.R3 * 0.5
        l_width = self.layout['width']
        l_height = self.layout['height']
        w_a, h_a = self.img_adj()
        img_w = l_width * tri_w - w_a
        img_h = l_height * tri_h - h_a
        pix_w = int(img_w)
        pix_h = int(img_h)
        return pix_w, pix_h

    def dim_from_image(self, pix_w: int, pix_h: int):
        """given the image dimensions, return the side of a triangle in pixels"""
        img_w = float(pix_w)
        img_h = float(pix_h)
        tri_h = self.R3 * 0.5
        tri_w = 1.0
        l_width = self.layout['width']
        l_height = self.layout['height']
        w_a, h_a = self.img_adj()
        img_w += w_a
        img_h += h_a
        img_w /= l_width * tri_w
        img_h /= l_height * tri_h
        return np.rint((img_w + img_h) / 2.0)

    def valid(self, pts: NDArray) -> NDArray:
        """
        Test that points are valid
        """
        if pts.shape[-1] < 2:
            raise ValueError('Points must have 2 dimensions')
        signs = self.pt_face(pts)
        return np.any(signs != 0, axis=1)

    def pt_face(self, pts: NDArray) -> NDArray:
        """Vectorised: identify octant sign for each point in net coordinates.
        Returns (N,3) int8 array of signs (±1), or (0,0,0) for invalid.
        """
        num_points = pts.shape[0]
        out = np.zeros((num_points, 3), dtype=np.int8)
        for sign, tri in self.face_tris.items():
            mask = inside_triangle_cw(pts, tri)
            if not np.any(mask):
                continue
            out[mask] = np.array(sign, dtype=np.int8)
        return out

    def px_pt(self, x, y, pix):
        """
        Given a pixel coordinate and the side of a
        triangle in pixels return the pt in this domain
        tri_w = pixels
        tri_h = pixels * self.R3 * 0.5
        """
        l_width = self.layout['width']
        l_height = self.layout['height']
        tri_w = pix
        tri_h = pix * H9K.derived.RH
        img_w = l_width * tri_w - (l_width + 0.51)
        img_h = l_height * tri_h - (l_height + 0.51)
        ux = self.wi * x/img_w
        uy = self.he * y/img_h
        oc = self.pt_face(np.array([ux, uy]))
        return oc, ux, uy

    def binning(self, pts: Points, sig: tuple = None):
        """Identify the components of the points"""
        cmp = self.pt_face(pts.coords)
        pts.components = np.array(cmp)

    def register_format(self, af: PointFormat):
        """Decorator to register an AddressFormat for each component."""
        for side in self.sides:
            self.sides[side].register_format(af)



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Experimental
uregion.py differs from region.py by method - using lattice+residual methods.
This is currently unsupported as it lacks the additional controls
offered by xy_regions, etc.
"""
from dataclasses import dataclass
from typing import Literal
import numpy as np
from numpy.typing import NDArray
from hhg9.h9.region import H9Context, H9CTX, StepEvent


@dataclass(slots=True)
class StepEventUV(StepEvent):
    space: Literal['uv'] = 'uv'


def xyu_regions_iter(xy, mode=None, depth=36, ctx: H9Context = H9CTX):
    """Generator mirroring xyu_regions but yielding per-layer StepEventUV.
    Yields two events per layer (pre/post) and returns the final addresses at StopIteration.
    """
    import hhg9.h9.classifier as clf
    h9k, h9cl, h9c, cr = ctx.k, ctx.cl, ctx.c, ctx.r
    invalid_region = cr.invalid_region
    proto_up, proto_dn = cr.proto_up, cr.proto_dn
    r3 = h9k.radical.R3
    r6 = np.sqrt(6.0)
    ẋ0 = xy[:, 0] * r3
    xu = ẋ0 * r6
    yv = 3.0 * xy[:, 1] * r6
    xu = snap_int(xu, k=3.5)
    yv = snap_int(yv, k=1.5)
    off_uv_f = h9c.off_uv.astype(np.float64, copy=False)
    addresses = np.full((xy.shape[0], depth + 2), invalid_region, dtype=np.uint8)
    addresses[:, 0] = np.where(mode == 1, proto_up, proto_dn)
    for i in range(depth + 1):
        cid = clf.classify_cell_uv(xu, yv)
        bad = ~cr.is_in[cid]
        yield StepEventUV(i=i, phase='pre', addresses=addresses, cid=cid, bad=bad, y=yv, ẋ=xu, pmo=None)
        if np.any(bad):
            # Pass 1: snap V
            yv[bad] = np.rint(yv[bad]).astype(yv.dtype, copy=False)
            cid_bad = clf.classify_cell_uv(xu[bad], yv[bad])
            still_bad = ~cr.is_in[cid_bad]
            # Pass 2: snap U
            if np.any(still_bad):
                # idx_sb = np.nonzero(still_bad)[0]
                xu_sub = xu[bad][still_bad]
                xu_sub = np.rint(xu_sub).astype(xu.dtype, copy=False)
                cid_bad2 = clf.classify_cell_uv(xu_sub, yv[bad][still_bad])
                cid_bad[still_bad] = cid_bad2
                xu[bad][still_bad] = xu_sub
                still_bad = ~cr.is_in[cid_bad]
            # Pass 3: metric fallback
            if np.any(still_bad):
                # Final fallback: choose the closest child in the CURRENT supercell (UV, float64)
                pm_bd = h9c.mode[addresses[bad, i]]
                pm_sb = pm_bd[still_bad]
                cand_mat = np.stack([cr.downs, cr.ups], axis=0)
                cand_sb = cand_mat[pm_sb]
                off_cand = off_uv_f[cand_sb]
                uv_sb = np.stack([xu[bad][still_bad], yv[bad][still_bad]], axis=1)
                d = off_cand - uv_sb[:, None, :]
                d2 = d[..., 0]*d[..., 0] + d[..., 1]*d[..., 1]
                pick = np.argmin(d2, axis=1)
                chosen = cand_sb[np.arange(pick.size), pick]
                cid_bad[np.nonzero(still_bad)[0]] = chosen
                still_bad = np.zeros_like(still_bad, dtype=bool)
            # Merge
            cid[bad] = cid_bad
            bad_next = np.zeros_like(bad)
            bad_next[bad] = still_bad
            bad = bad_next

        good = ~bad
        if np.any(good):
            g_id = cid[good]
            uvo_g = off_uv_f[g_id]
            xu[good] = 3.0 * xu[good] - 3.0 * uvo_g[:, 0]
            yv[good] = 3.0 * yv[good] - 3.0 * uvo_g[:, 1]
            addresses[good, i + 1] = g_id

        yield StepEventUV(i=i, phase='post', addresses=addresses, cid=cid, bad=bad, y=yv, ẋ=xu)
    return addresses


def snap_int(z, k: float = 1.0, ulp_floor: float | None = None):
    """Snap values that are within ~k ULPs of an integer to that integer."""
    zi = np.rint(z)
    diff = np.abs(z - zi)
    # per-element ULP at the integer (distance to next float)
    ulp = np.abs(np.nextafter(zi, np.inf) - zi)
    if ulp_floor is None:
        ulp_floor = np.finfo(z.dtype).eps  # ~2.22e-16 for float64
    ulp = np.maximum(ulp, ulp_floor)
    max_diff = k * ulp
    mask = diff <= max_diff
    return np.where(mask, zi, z)


def xyu_regions(xy, mode=None, depth=36, ctx: H9Context = None) -> NDArray[np.uint8]:
    """Calculate regions via UV array - more integer math, mainly handling residuals"""
    # Use the iterator but materialize the result for compatibility.
    it = xyu_regions_iter(xy, mode=mode, depth=depth, ctx=ctx)
    addresses = None
    for ev in it:
        pass
    addresses = ev.addresses if 'ev' in locals() else None
    return addresses



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
find_coords is a general root finding algorithm for finding coordinates
on the ellipsoidal surface of an equilateral polyhedron projection.
"""
import numpy as np
from hhg9.h9.protocols import H9CellLike


def find_coords(target_rll, initial_mode, target_octants, h9c: H9CellLike,
                projector_func, distance_func, depth=34, beam_width=5):
    """
    Vectorised beam search in barycentric xy space.
    target_rll: (N,2) lat/lon (or whatever the projector uses)
    initial_mode: (N,) int {0,1}
    target_octants: (N,3) face-signs for projection
    h9c: H9Cell instance providing mode, offsets, and child region info
    projector_func: (XY,(...))->(M,2) projects bary XY to target space
    distance_func: ((M,2),(M,2))->(M,) distances in target space
    offset_kind: which offset flavour to use ("xy", "xbar_y"/"ẋy"/"xby", or "uv")
    """
    num_pts = target_rll.shape[0]  # N
    off = h9c.off_xy

    # child region id lists per supercell mode (shape (9,))
    up_children = h9c.ups
    down_children = h9c.downs
    mode_lut = h9c.mode

    # Roots by mode: up→0x16, down→0x49
    root_uris = np.where(initial_mode == 1, 0x16, 0x49).astype(np.uint8)

    # Beam state
    best_paths = root_uris[:, None, None]  # (N,1,1)
    best_xy = np.zeros((num_pts, 1, 2), dtype=np.float64)  # (N,1,2)

    for i in range(depth):
        k = best_paths.shape[1]  # current beam width
        last = best_paths[:, :, -1]  # (N,k)
        par_mode = mode_lut[last]  # (N,k)

        # Select children per parent mode
        children = np.where(par_mode[..., None] == 1, up_children, down_children)  # (N,k,9)

        # Expand candidate URIs and coords
        cand_uris = children.reshape(num_pts, k * 9)  # (N,k*9)
        scale = (1.0 / 3.0) ** i
        par_xy = np.repeat(best_xy, 9, axis=1)  # (N,k*9,2)
        offs = off[cand_uris]  # (N,k*9,2)
        cand_xy = par_xy + offs * scale  # (N,k*9,2)

        # Project once: flatten → project → reshape
        nk9 = num_pts * (k * 9)
        flat_xy = cand_xy.reshape(nk9, 2)
        flat_oct = np.repeat(target_octants, k * 9, axis=0)  # (nk9,3)
        proj = projector_func(flat_xy, flat_oct).reshape(num_pts, k * 9, 2)

        # Distances to targets
        d = distance_func(proj, target_rll[:, None, :])  # (N,k*9)

        # Top-k via arg-partition, then stable tie-break by URI
        idx_k = np.argpartition(d, beam_width - 1, axis=1)[:, :beam_width]  # (N,k)
        # Gather those distances/uris for stable sort
        d_k = np.take_along_axis(d, idx_k, axis=1)  # (N,k)
        u_k = np.take_along_axis(cand_uris, idx_k, axis=1)  # (N,k)
        # Stable rank on (distance, uri)
        order = np.lexsort((u_k, d_k))  # (N,k)
        top_idx = np.take_along_axis(idx_k, order, axis=1)  # (N,k)

        # Keep winners
        win_uris = np.take_along_axis(cand_uris, top_idx, axis=1)  # (N,k)
        win_xy = np.take_along_axis(cand_xy, top_idx[..., None], axis=1)  # (N,k,2)

        # Extend paths: pick corresponding parent slices
        parent_sel = top_idx // 9  # (N,k)
        parent_paths = np.take_along_axis(best_paths, parent_sel[..., None], axis=1)  # (N,k,depth_so_far)
        best_paths = np.concatenate([parent_paths, win_uris[..., None]], axis=2)  # (N,k,depth_so_far+1)
        best_xy = win_xy

    # Take the first beam entry as the result
    return best_xy[:, 0, :], best_paths[:, 0, :]



"""
Unit tests for the grid region encoding function en_grid.
Validates correct region classification based on geometric thresholds.
"""
from functools import lru_cache, cache
import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_allclose


@pytest.fixture
def const_grid():
    """Fixture providing a configured instance of RegionClassificationGrid."""
    return GridConstants()


class GridConstants:
    """A minimal class that only holds geometric constants and those methods that use only those constants."""

    def __init__(self):
        self.H = np.sqrt(6) / 2.
        self.R3 = np.sqrt(3)
        # All other geometric constants derived from H and R3
        # Done in order to avoid tiny floating point deviations.
        self.TR = self.H / self.R3  #
        self.W = 2 * self.TR  # This correctly derives W = sqrt(2)
        self.ΛC = 2 * self.H / 3.
        self.ΛF = -self.H / 3.
        self.VC = self.H / 3.
        self.VF = -2 * self.H / 3.
        self.Ẇ = self.ΛC
        self.TL = -self.TR
        self.U, self.V = self.W / 6., self.H / 9.

    def region_classification(self, ẋ, y):
        """
        Classify coordinates into geometric grid regions using thresholds.
        :param ẋ: np_array of √3-scaled x coordinates.
        :param y: np_array of y coordinates.
        :return: np_array of encoded region identifiers.
        """
        h_conditions = [
            y > self.ΛC,
            y > self.VC,
            y > 0,
            y > self.ΛF,
            y >= self.VF,
        ]
        h_id = np.select(h_conditions, [0, 1, 2, 3, 4], default=5)
        y_minus_x = y - ẋ
        p_conditions = [
            y_minus_x > self.Ẇ,
            y_minus_x > 0,
            y_minus_x >= -self.Ẇ,
        ]
        p_id = np.select(p_conditions, [0, 1, 2], default=3)
        y_plus_x = y + ẋ
        n_conditions = [
            y_plus_x < -self.Ẇ,
            y_plus_x < 0,
            y_plus_x <= self.Ẇ,
        ]
        n_id = np.select(n_conditions, [0, 1, 2], default=3)
        return h_id << 4 | p_id << 2 | n_id


@pytest.fixture
def reg_grid():
    """Fixture providing an instance of GridRegions."""
    return GridRegions()


class GridRegions(GridConstants):
    def __init__(self):
        super().__init__()
        self.num_regions = 96    # This number follows from the partitions within GridConstants
        self.invalid_ugc = 0x5f  # (final region := OOB)

        # Build offsets analytically (exact integers × base units)
        hx3 = self.H / 3.0  # unit for ẋ offsets
        hy9 = self.H / 9.0  # unit for y offsets

        # The index order is such that in_regions [0..5] are mode-shared: (found inside both modes),
        # and index % 2 indicates the mode of that region.
        # Mode specific regions are found in those regions which have the same mode as themselves.
        self.in_regions = [
            0x26, 0x2a, 0x3a, 0x39, 0x35, 0x25,  # Mode Shared, inner regions from V above barycentric origin, CW.
            0x49, 0x34, 0x21, 0x16, 0x2b, 0x3e   # Mode Specific outer regions from apex of V, CW
        ]
        self.in_up_regions = [  # 9 regions serve Λ mode
            0x39, 0x3a, 0x3e,  # c0 ΛVΛ
            0x25, 0x35, 0x34,  # c1 ΛVΛ
            0x2a, 0x26, 0x16,  # c2 ΛVΛ
        ]
        self.in_dn_regions = [
            0x26, 0x2a, 0x2b,  # c0 VΛV
            0x3a, 0x39, 0x49,  # c1 VΛV
            0x35, 0x25, 0x21,  # c2 VΛV
        ]

        # origin offset multipliers corresponding to `in_regions`:
        mx = np.array([
            0, +1, +1,  # 0x26, 0x2a, 0x3a
            0, -1, -1,  # 0x39, 0x35, 0x25
            0, -2, -2,  # 0x49, 0x34, 0x21
            0, +2, +2,  # 0x16, 0x2b, 0x3e
        ], dtype=np.int8)

        my = np.array([
            +2, +1, -1,  # 0x26, 0x2a, 0x3a
            -2, -1, +1,  # 0x39, 0x35, 0x25
            -4, -2, +2,  # 0x49, 0x34, 0x21
            +4, +2, -2,  # 0x16, 0x2b, 0x3e
        ], dtype=np.int8)

        self.ugc_off_x = np.zeros(self.num_regions, dtype=np.float64)
        self.ugc_off_y = np.zeros(self.num_regions, dtype=np.float64)
        # self.ugc_off_ẋ = np.zeros(self.num_regions, dtype=np.float64)

        self.ugc_off_x[self.in_regions] = mx * hx3
        self.ugc_off_y[self.in_regions] = my * hy9
        self.ugc_off = np.hstack((self.ugc_off_x, self.ugc_off_y))
        # self.ugc_off_ẋ[self.in_regions] = mx * hx3 * self.R3

        ugc_num_props = 3

        (self.in_dn, self.in_up, self.mode) = range(ugc_num_props)
        self.ugc_lut = np.full((self.num_regions, ugc_num_props), self.invalid_ugc, dtype=np.uint8)
        self.ugc_lut[:, self.in_dn] = 0
        self.ugc_lut[:, self.in_up] = 0
        self.ugc_lut[self.in_up_regions, self.in_up] = 1
        self.ugc_lut[self.in_dn_regions, self.in_dn] = 1
        self.ugc_lut[self.in_regions, self.mode] = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
        self.ugc_lut[self.invalid_ugc] = 0  # set no offsets for illegal!


    @cache
    def child_lut(self):
        """
        Return LUT which, given a mode and c1, returns the child regions there.
        The first two children are mode-shared.
        The middle switches mode from parent.
        The final mode is mode-locked.
        Necessary feature is that a child-region-c1 is unique.
        """
        _chd = {
            (0, 0): [0x26, 0x2A, 0x2B],  # V,C1.0 VΛV
            (0, 1): [0x3A, 0x39, 0x49],  # V,C1.1 VΛV
            (0, 2): [0x35, 0x25, 0x21],  # V,C1.2 VΛV
            (1, 0): [0x39, 0x3A, 0x3E],  # Λ,C1.0 ΛVΛ
            (1, 1): [0x25, 0x35, 0x34],  # Λ,C1.1 ΛVΛ
            (1, 2): [0x2A, 0x26, 0x16],  # Λ,C1.2 ΛVΛ
        }
        lut = np.zeros((2, 3, 3), dtype=np.uint8)
        for (mode, c1), children in _chd.items():
            lut[mode, c1] = children
        return lut

    @cache
    def pc_c1(self):
        """
        Build (parent_region, child_region) -> c1 from canonical child_lut + ugc_lut.
        """
        num = self.num_regions
        lut = np.full((num, num), self.invalid_ugc, dtype=np.uint8)
        # lut = np.zeros((num, num), dtype=np.uint8)

        ugc = self.ugc_lut  # (num_regions, 1) -> mode
        chd = self.child_lut()  # (2, 3, 3)

        # For every parent region p, infer its mode and fill reverse mapping.
        for par in self.in_regions:
            mode = ugc[par, self.mode]
            for c1 in (0, 1, 2):
                for child in chd[mode, c1]:
                    lut[par, child] = c1
        return lut

    # def ugc_regions(self, x, y, mode, depth=36):
    #     """
    #     Given a vector of Point coords create a set of regions
    #     """
    #     num_points = x.size
    #     addresses = np.full((num_points, depth + 2), self.invalid_ugc, dtype=np.uint8)
    #     addresses[:, 0] = np.where(mode == 1, 0x16, 0x49)  # These values should come from the octant set.
    #     # history = np.zeros((num_points, depth + 2, 6))
    #     for i in range(depth + 1):
    #         ẋ = self.R3 * x
    #         region = self.region_classification(ẋ, y)  # Raw classification
    #         props = self.ugc_lut[region]
    #         mode_up = props[:, self.in_up]
    #         mode_dn = props[:, self.in_dn]
    #         in_scope = np.where(mode == 1, mode_up, mode_dn)
    #         region_id = np.where(in_scope, region, self.invalid_ugc)  # Validated ID
    #         addresses[:, i + 1] = region_id
    #         off = self.ugc_off[region_id]
    #         mode = self.ugc_lut[region_id, self.mode]
    #         x -= off[:, 0]
    #         y -= off[:, 1]
    #         x *= 3.
    #         y *= 3.
    #     return addresses

    def ugc_regions(self, xy, mode, depth=36):
        """
        Given a vector of Point coords create a set of regions
        """
        num_points = xy.shape[0]
        x = np.copy(xy[:, 0])
        y = np.copy(xy[:, 1])
        addresses = np.full((num_points, depth + 2), self.invalid_ugc, dtype=np.uint8)
        addresses[:, 0] = np.where(mode == 1, 0x16, 0x49)  # These values should come from the octant set.
        for i in range(depth + 1):
            ẋ = self.R3 * x
            region = self.region_classification(ẋ, y)  # Raw classification
            props = self.ugc_lut[region]
            mode_up = props[:, self.in_up]
            mode_dn = props[:, self.in_dn]
            in_scope = np.where(mode == 1, mode_up, mode_dn)
            region_id = np.where(in_scope, region, self.invalid_ugc)  # Validated ID
            addresses[:, i + 1] = region_id
            off = self.ugc_off[region_id]
            mode = self.ugc_lut[region_id, self.mode]
            x -= off[:, 0]
            y -= off[:, 1]
            x *= 3.
            y *= 3.
        return addresses


    def ugc_dec(self, uri_address):
        """
        Decode back to (x,y,mode) by undoing the classifier-plane stepping.
        """
        num_points, depth = uri_address.shape
        ẋ = np.zeros(num_points, dtype=np.float64)
        y = np.zeros(num_points, dtype=np.float64)

        for i in range(depth - 2, 0, -1):
            region_id = uri_address[:, i]
            valid = (region_id != self.invalid_ugc)

            ẋ /= 3.0
            y /= 3.0

            if np.any(valid):
                rid = region_id[valid]
                ẋ[valid] += self.ugc_off_ẋ[rid]
                y[valid] += self.ugc_off_y[rid]

        initial_mode = self.ugc_lut[uri_address[:, 0], self.mode]
        x = ẋ / self.R3
        return np.stack([x, y, initial_mode], axis=-1)

    def ugc_regions_w(self, xy, mode, depth=36):
        """
        Given a vector of Point coords create a set of regions
        """
        num_points = xy.shape[0]
        x = np.copy(xy[:, 0])
        y = np.copy(xy[:, 1])
        addresses = np.full((num_points, depth + 2), self.invalid_ugc, dtype=np.uint8)
        addresses[:, 0] = np.where(mode == 1, 0x16, 0x49)  # These values should come from the octant set.
        for i in range(depth + 1):
            # x, y, ẋ = self.clamp(x, y, mode)
            ẋ = self.R3 * x
            region = self.region_classification(ẋ, y)  # Raw classification
            props = self.ugc_lut[region]
            mode_up = props[:, self.in_up]
            mode_dn = props[:, self.in_dn]
            in_scope = np.where(mode == 1, mode_up, mode_dn)
            region_id = np.where(in_scope, region, self.invalid_ugc)  # Validated ID
            addresses[:, i + 1] = region_id
            mode = self.ugc_lut[region_id, self.mode]
            x -= self.ugc_off_x[region_id]
            y -= self.ugc_off_y[region_id]
            x *= 3.
            y *= 3.
        return self.terminate(addresses)

    def ugc_dec_w(self, uri_address):
        """
        REVERSE: URI addresses back into (x,y) coordinates and initial mode.
        Inverse of ugc_regions
        """
        num_points, depth = uri_address.shape
        # Initialize x and y with the precise remainder from the encoding process.
        x = np.zeros(num_points, dtype=np.float64)
        y = np.zeros(num_points, dtype=np.float64)

        # Loop backwards from the penultimate layer down to the first REAL layer (index 1),
        # skipping the placeholder root at index 0.
        # Likewise, leaf residual is ignored by design; its mag. is ≤ ~7·3⁻ᵏ and below fp noise for k≥34.
        for i in range(depth - 2, 0, -1):
            region_id = uri_address[:, i]
            valid_mask = (region_id != self.invalid_ugc)

            x /= 3.0
            y /= 3.0

            if np.any(valid_mask):
                valid_ids = region_id[valid_mask]
                x[valid_mask] += self.ugc_off_x[valid_ids]
                y[valid_mask] += self.ugc_off_y[valid_ids]

        # After reconstructing the coordinates, find the initial mode from the root URI.
        initial_mode = self.ugc_lut[uri_address[:, 0], self.mode]

        # Stack all three results into a final (N, 3) array.
        return np.stack([x, y, initial_mode], axis=-1)

    def alt_ugc_dec(self, uri_address):
        """
        REVERSE: Decodes a URI back into (x,y) and initial mode.
        Runs backward in (ẋ, y); converts ẋ→x only at the end.
        """
        num_points, depth = uri_address.shape
        ẋ = np.zeros(num_points, dtype=np.float64)
        y = np.zeros(num_points, dtype=np.float64)

        for i in range(depth - 1, 0, -1):
            region_id = uri_address[:, i]
            valid = (region_id != self.invalid_ugc)

            ẋ /= 3.0
            y /= 3.0

            if np.any(valid):
                rid = region_id[valid]
                ẋ[valid] += self.ugc_off_ẋ[rid]
                y[valid] += self.ugc_off_y[rid]

        initial_mode = self.ugc_lut[uri_address[:, 0], self.mode]
        x = ẋ / self.R3  # convert back once
        return np.stack([x, y, initial_mode], axis=-1)

    def alt_ugc_regions(self, x, y, mode, depth=36):
        num_points = x.size
        addresses = np.full((num_points, depth + 2), self.invalid_ugc, dtype=np.uint8)
        addresses[:, 0] = np.where(mode == 1, 0x16, 0x49)

        for i in range(depth + 1):
            # 1) classify in classifier plane
            xbar = self.R3 * x
            region = self.region_classification(xbar, y)
            props = self.ugc_lut[region]

            # 2) validate region vs parent mode (in-up / in-down)
            in_scope = np.where(mode == 1, props[:, self.in_up], props[:, self.in_dn])
            region_id = np.where(in_scope, region, self.invalid_ugc)
            addresses[:, i + 1] = region_id

            # 3) step one layer **in the classifier plane** for x, native for y;
            #    then convert back to x for the next iteration
            off_ẋ = self.ugc_off_ẋ[region_id]  # √3-scaled x-offsets
            off_y = self.ugc_off_y[region_id]
            mode = self.ugc_lut[region_id, self.mode]

            xbar = (xbar - off_ẋ) * 3.0
            y = (y - off_y) * 3.0
            x = xbar / self.R3

        return addresses

    def trace_roundtrip(self, xy, mode, depth, name):
        addr = self.ugc_regions(xy, mode, depth=depth)
        ugc = self.ugc_lut
        pc = self.pc_c1()
        chd = self.child_lut()

        # ----- forward (encode) -----
        x = xy[:, 0].copy()
        y = xy[:, 1].copy()
        x, y, xbar = self.clamp(x, y, mode)  # encoder’s very first step
        f = []
        f.append(dict(i=0, par=addr[0, 0], cur=None, mo=int(mode[0]),
                      x=float(x), y=float(y), xbar=float(xbar)))
        for i in range(1, depth + 1):
            par = addr[0, i - 1]
            cur = addr[0, i]
            c1 = pc[par, cur]
            mo = ugc[par, self.mode]
            offx = self.ugc_off_ẋ[cur]
            offy = self.ugc_off_y[cur]
            # step
            xbar2 = (xbar - offx) * 3.0
            y2 = (y - offy) * 3.0
            x2 = xbar2 / self.R3
            f.append(dict(i=i, par=int(par), cur=int(cur), c1=int(c1), mo=int(mo),
                          offx=float(offx), offy=float(offy),
                          x=float(x2), y=float(y2), xbar=float(xbar2)))
            x, y, xbar = x2, y2, xbar2

        # ----- backward (decode) -----
        bxbar = 0.0
        by = 0.0
        b = []
        for i in range(depth, 0, -1):
            rid = addr[0, i]
            bxbar = bxbar / 3.0 + float(self.ugc_off_ẋ[rid])
            by = by / 3.0 + float(self.ugc_off_y[rid])
            b.append(dict(i=i, rid=int(rid), x=float(bxbar / self.R3), y=float(by), xbar=float(bxbar)))
        b = list(reversed(b))  # i asc

        # ----- compare per step -----
        # Precompute root-frame totals from decode
        x0_bar = b[-1]['xbar']  # total \bar x_0
        y0 = b[-1]['y']  # total y_0

        # Also precompute prefix sums of off/3^(k-1)
        offx = [float(self.ugc_off_ẋ[addr[0, k]]) for k in range(1, depth + 1)]
        offy = [float(self.ugc_off_y[addr[0, k]]) for k in range(1, depth + 1)]
        px = 0.0
        py = 0.0

        print(f"\nTRACE {name} depth={depth}")
        for i in range(0, depth + 1):
            if i == 0:
                # Encoder's initial clamped state vs decoder's reconstructed root state
                dec_x = x0_bar / self.R3
                dec_y = y0
                dx = f[i]['x'] - dec_x
                dy = f[i]['y'] - dec_y
                print(f" i={i:02d} enc(x,y)=({f[i]['x']:+.9f},{f[i]['y']:+.9f}) "
                      f"dec_root=({dec_x:+.9f},{dec_y:+.9f}) Δ=({dx:+.3e},{dy:+.3e})")
                continue

            # Update prefix sums up to i
            px = px + offx[i - 1] / (3.0 ** (i - 1))
            py = py + offy[i - 1] / (3.0 ** (i - 1))

            # Decode-side prediction in the SAME frame as enc step i
            dec_xbar_i = (3.0 ** i) * (x0_bar - px)
            dec_y_i = (3.0 ** i) * (y0 - py)
            dec_x_i = dec_xbar_i / self.R3

            dx = f[i]['x'] - dec_x_i
            dy = f[i]['y'] - dec_y_i

            print(
                f" i={i:02d} par=0x{f[i]['par']:02x} cur=0x{f[i]['cur']:02x} c1={f[i]['c1']} mo={f[i]['mo']} "
                f"enc=({f[i]['x']:+.9f},{f[i]['y']:+.9f}) "
                f"dec_pred=({dec_x_i:+.9f},{dec_y_i:+.9f}) Δ=({dx:+.3e},{dy:+.3e})"
            )

    def diag_first_step(self, xy, mode, depth=8, name="case"):
        reg = self
        addr = reg.ugc_regions(xy, mode, depth=depth)
        ugc  = reg.ugc_lut

        # 1) Encoder’s clamped root (what the classifier actually used)
        x, y, xbar = reg.clamp(xy[:,0].copy(), xy[:,1].copy(), mode)
        par = int(addr[0,0])
        cur = int(addr[0,1])
        mo  = int(ugc[par, reg.mode])

        print(f"\n{ name }")
        print(f" par=0x{par:02x} mo={mo} chosen cur=0x{cur:02x}")
        print(f" enc clamp root: xbar0={xbar[0]:+.12f}  y0={y[0]:+.12f}")

        # 2) Decode-implied root from the whole address (unique series sum)
        offx = reg.ugc_off_ẋ[addr[0,1:depth+1]].astype(np.float64)
        offy = reg.ugc_off_y[addr[0,1:depth+1]].astype(np.float64)
        pow3 = (3.0 ** np.arange(depth, 0, -1))
        xbar0_dec = np.sum(offx / pow3)
        y0_dec    = np.sum(offy / pow3)
        print(f" dec series root: xbar0={xbar0_dec:+.12f}  y0={y0_dec:+.12f}")
        print(f" Δ root: dx={xbar[0]-xbar0_dec:+.12e}  dy={y[0]-y0_dec:+.12e}")

        # 3) What would the classifier pick at the decode root?
        rid_pred = reg.region_classification(
            np.array([xbar0_dec]), np.array([y0_dec])
        )[0]
        print(f" classifier at dec-root ⇒ rid_pred=0x{rid_pred:02x}  (vs chosen cur=0x{cur:02x})")

        return par, mo, cur, rid_pred, (xbar[0]-xbar0_dec), (y[0]-y0_dec)


@pytest.fixture
def rel_grid():
    """Fixture providing an instance of GridNeighbours."""
    return GridNeighbours()


class GridNeighbours:
    """
        Class that provides enough to test neighbour function.
    """

    def __init__(self):
        self.invalid_ugc = 0x5f
        self.num_regions = 96
        self.ugc_num_props = 1  # for relations, we only need mode metadata
        (self.mode,) = range(self.ugc_num_props)  # indices of metadata

    @cache
    def ugc_lut(self):
        """
        UGC Metadata - Here it is going to just be the mode.
        When testing neighbours we will only need 1 child.
        """
        num_regions = self.num_regions
        # these regions are ordered - eg, self.ugc_lut[self.in_regions, self.mode]
        _in_regions = [
            0x26, 0x2a, 0x3a, 0x39, 0x35, 0x25,  # Mode Shared, inner regions from V above barycentric origin, CW.
            0x49, 0x34, 0x21, 0x16, 0x2b, 0x3e   # Mode Specific outer regions from apex of V, CW
        ]
        _in_up_regions = [  # 9 regions serve Λ mode
            0x39, 0x3a, 0x3e,  # c0 ΛVΛ
            0x25, 0x35, 0x34,  # c1 ΛVΛ
            0x2a, 0x26, 0x16,  # c2 ΛVΛ
        ]
        _in_dn_regions = [
            0x26, 0x2a, 0x2b,  # c0 VΛV
            0x3a, 0x39, 0x49,  # c1 VΛV
            0x35, 0x25, 0x21,  # c2 VΛV
        ]
        _ugc_lut = np.full((num_regions, self.ugc_num_props), self.invalid_ugc, dtype=np.uint8)
        _ugc_lut[_in_regions, self.mode] = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
        return _ugc_lut

    @cache
    def child_lut(self):
        """
        Given a mode and c1, find the regions that belong there.
        When testing neighbours we will only need 1 child.
        """
        _chd = {
            (0, 0): [0x26, 0x2A, 0x2B],  # V,C1.0
            (0, 1): [0x3A, 0x39, 0x49],  # V,C1.1
            (0, 2): [0x35, 0x25, 0x21],  # V,C1.2
            (1, 0): [0x39, 0x3A, 0x3E],  # Λ,C1.0
            (1, 1): [0x25, 0x35, 0x34],  # Λ,C1.1
            (1, 2): [0x2A, 0x26, 0x16],  # Λ,C1.2
        }
        _child_lut = np.zeros((2, 3, 3), dtype=np.uint8)
        for (mode, c1), children in _chd.items():
            _child_lut[mode, c1] = children
        return _child_lut

    @cache
    def pqc1_lut(self):
        """
        Given parent region, child region, return the C1.
        It's lazy, but it's easy to use.
        It uses @cache in order to be lazy-loaded and not to clutter up the __init__
        """
        _pqc1 = {
            (0x21, 0x26): 0, (0x21, 0x2a): 0, (0x21, 0x2b): 0,
            (0x21, 0x3a): 1, (0x21, 0x39): 1, (0x21, 0x49): 1,
            (0x21, 0x35): 2, (0x21, 0x25): 2, (0x21, 0x21): 2,
            (0x26, 0x26): 0, (0x26, 0x2a): 0, (0x26, 0x2b): 0,
            (0x26, 0x3a): 1, (0x26, 0x39): 1, (0x26, 0x49): 1,
            (0x26, 0x35): 2, (0x26, 0x25): 2, (0x26, 0x21): 2,
            (0x2b, 0x26): 0, (0x2b, 0x2a): 0, (0x2b, 0x2b): 0,
            (0x2b, 0x3a): 1, (0x2b, 0x39): 1, (0x2b, 0x49): 1,
            (0x2b, 0x35): 2, (0x2b, 0x25): 2, (0x2b, 0x21): 2,
            (0x35, 0x26): 0, (0x35, 0x2a): 0, (0x35, 0x2b): 0,
            (0x35, 0x3a): 1, (0x35, 0x39): 1, (0x35, 0x49): 1,
            (0x35, 0x35): 2, (0x35, 0x25): 2, (0x35, 0x21): 2,
            (0x3a, 0x26): 0, (0x3a, 0x2a): 0, (0x3a, 0x2b): 0,
            (0x3a, 0x3a): 1, (0x3a, 0x39): 1, (0x3a, 0x49): 1,
            (0x3a, 0x35): 2, (0x3a, 0x25): 2, (0x3a, 0x21): 2,
            (0x49, 0x26): 0, (0x49, 0x2a): 0, (0x49, 0x2b): 0,
            (0x49, 0x3a): 1, (0x49, 0x39): 1, (0x49, 0x49): 1,
            (0x49, 0x35): 2, (0x49, 0x25): 2, (0x49, 0x21): 2,
            (0x16, 0x39): 0, (0x16, 0x3a): 0, (0x16, 0x3e): 0,
            (0x16, 0x25): 1, (0x16, 0x35): 1, (0x16, 0x34): 1,
            (0x16, 0x2a): 2, (0x16, 0x26): 2, (0x16, 0x16): 2,
            (0x25, 0x39): 0, (0x25, 0x3a): 0, (0x25, 0x3e): 0,
            (0x25, 0x25): 1, (0x25, 0x35): 1, (0x25, 0x34): 1,
            (0x25, 0x2a): 2, (0x25, 0x26): 2, (0x25, 0x16): 2,
            (0x2a, 0x39): 0, (0x2a, 0x3a): 0, (0x2a, 0x3e): 0,
            (0x2a, 0x25): 1, (0x2a, 0x35): 1, (0x2a, 0x34): 1,
            (0x2a, 0x2a): 2, (0x2a, 0x26): 2, (0x2a, 0x16): 2,
            (0x34, 0x39): 0, (0x34, 0x3a): 0, (0x34, 0x3e): 0,
            (0x34, 0x25): 1, (0x34, 0x35): 1, (0x34, 0x34): 1,
            (0x34, 0x2a): 2, (0x34, 0x26): 2, (0x34, 0x16): 2,
            (0x39, 0x39): 0, (0x39, 0x3a): 0, (0x39, 0x3e): 0,
            (0x39, 0x25): 1, (0x39, 0x35): 1, (0x39, 0x34): 1,
            (0x39, 0x2a): 2, (0x39, 0x26): 2, (0x39, 0x16): 2,
            (0x3e, 0x39): 0, (0x3e, 0x3a): 0, (0x3e, 0x3e): 0,
            (0x3e, 0x25): 1, (0x3e, 0x35): 1, (0x3e, 0x34): 1,
            (0x3e, 0x2a): 2, (0x3e, 0x26): 2, (0x3e, 0x16): 2,
        }
        _pqc1_lut = np.zeros((self.num_regions, self.num_regions), dtype=np.uint8)
        for (p_reg, c_reg), c1 in _pqc1.items():
            _pqc1_lut[p_reg, c_reg] = c1
        return _pqc1_lut

    @lru_cache(maxsize=None)
    def neighbour_lut(self):
        """
        Given a region, parent mode, region-c1, return the neighbour and parent mode.
        If the parent mode has changed, then the region parent is a neighbour.
        It uses @lru_cache in order to be lazy-loaded and not to clutter up the __init__
        """
        _lut = {
            (0x16, 1): [(0x26, 1), (0x2B, 0), (0x21, 0)],
            (0x21, 1): [(0x5F, 1), (0x5F, 1), (0x5F, 1)],
            (0x25, 1): [(0x35, 1), (0x3A, 0), (0x26, 1)],
            (0x26, 1): [(0x16, 1), (0x2A, 1), (0x25, 1)],
            (0x2A, 1): [(0x3A, 1), (0x26, 1), (0x35, 0)],
            (0x2B, 1): [(0x5F, 1), (0x5F, 1), (0x5F, 1)],
            (0x34, 1): [(0x21, 0), (0x49, 0), (0x35, 1)],
            (0x35, 1): [(0x25, 1), (0x39, 1), (0x34, 1)],
            (0x39, 1): [(0x26, 0), (0x35, 1), (0x3A, 1)],
            (0x3A, 1): [(0x2A, 1), (0x3E, 1), (0x39, 1)],
            (0x3E, 1): [(0x2B, 0), (0x3A, 1), (0x49, 0)],
            (0x49, 1): [(0x5F, 1), (0x5F, 1), (0x5F, 1)],
            (0x16, 0): [(0x5F, 0), (0x5F, 0), (0x5F, 0)],
            (0x21, 0): [(0x34, 1), (0x25, 0), (0x16, 1)],
            (0x25, 0): [(0x35, 0), (0x21, 0), (0x26, 0)],
            (0x26, 0): [(0x39, 1), (0x2A, 0), (0x25, 0)],
            (0x2A, 0): [(0x3A, 0), (0x26, 0), (0x2B, 0)],
            (0x2B, 0): [(0x3E, 1), (0x16, 1), (0x2A, 0)],
            (0x34, 0): [(0x5F, 0), (0x5F, 0), (0x5F, 0)],
            (0x35, 0): [(0x25, 0), (0x39, 0), (0x2A, 1)],
            (0x39, 0): [(0x49, 0), (0x35, 0), (0x3A, 0)],
            (0x3A, 0): [(0x2A, 0), (0x25, 1), (0x39, 0)],
            (0x3E, 0): [(0x5F, 0), (0x5F, 0), (0x5F, 0)],
            (0x49, 0): [(0x39, 0), (0x34, 1), (0x3E, 1)],
        }
        _neighbour_lut = np.full((self.num_regions, 2, 3, 2), self.invalid_ugc, dtype=np.uint8)
        for key, neighbours in _lut.items():
            region_id, mode = key
            _neighbour_lut[region_id, mode] = neighbours
        return _neighbour_lut

    def region_neighbours(self, address):
        """
        Vectorised means to return neighbouring half-hexagon addresses (as regions) via regions.
        The last value-holding region (addresses[:, -2]) is the key position of interest (POI).
        But we must cascade when necessary also.
        """
        count, layers = address.shape
        neighbour = address.copy()  # The neighbour may just be a single switch.
        cascading = np.ones(count, dtype=bool)  # Track all the addresses we are managing.
        n_lut = self.neighbour_lut()
        c1_lut = self.pqc1_lut()
        ugc_lut = self.ugc_lut()
        c1 = c1_lut[address[:, -2], address[:, -1]]
        for poi in range(layers - 2, -1, -1):
            if not np.any(cascading):
                break
            active = np.where(cascading)[0]
            cur = address[:, poi][active]
            par = address[:, poi - 1][active]
            pmo = ugc_lut[par, self.mode]
            nbm = n_lut[cur, pmo, c1[active]]
            neighbour[:, poi][active] = nbm[:, -2]
            cascading[active] = (nbm[:, 1] != pmo)
        # Normalise root and terminal.
        nmo = ugc_lut[neighbour[:, 0], self.mode]
        root = np.where(nmo == 1, 0x16, 0x49)
        neighbour[:, 0] = root
        child_lut = self.child_lut()
        mode = ugc_lut[neighbour[:, -2], self.mode]  # mode of region.
        neighbour[:, -1] = child_lut[mode, c1, 2]
        return neighbour

    def encroach_to_neighbour(self, xy, mode, c1_edge, preserve):
        # xy: (N,2) local coords in source face
        # mode: (N,) source face mode (not used in math here, but good to thread)
        # c1_edge: 0,1,2 – which edge family you crossed (on the source face)
        # preserve: bool – whether this face-to-face seam preserves or swaps C1

        # 1) go to classifier plane
        xbar = np.sqrt(3) * xy[:, 0]
        vec = np.stack([xbar, xy[:, 1]], axis=-1)

        # 2) reflect if swap seam
        if not preserve:
            if c1_edge == 0:
                u = np.array([1.0, 0.0])
            elif c1_edge == 1:
                u = np.array([1.0, 1.0]) / np.sqrt(2)
            elif c1_edge == 2:
                u = np.array([1.0, -1.0]) / np.sqrt(2)
            R = 2 * np.outer(u, u) - np.eye(2)
            vec = vec @ R.T

        # 3) back to (x,y) in neighbour frame
        x_prime = vec[:, 0] / np.sqrt(3)
        y_prime = vec[:, 1]
        return np.stack([x_prime, y_prime], axis=-1)


# def test_clamp_up_mode(const_grid):
#     """Tests the clamping logic for the UP (Λ) triangle."""
#     grid = const_grid
#     test_cases = {
#         "Inside": (np.array([[0.1, 0.1]]), np.array([[0.1, 0.1]])),
#         "Outside Apex": (np.array([[0.1, 1.0]]), np.array([[0.0, grid.ΛC]])),
#         "Pole": (
#             np.array([[0.000000000000001, 0.8164965132120238]]), np.array([[0.000000000000001, 0.8164965132120238]])),
#         "Outside Right Slant": (np.array([[0.8, 0.2]]), np.array([[0.355934, 0.2]])),
#         "Outside Base-Left": (np.array([[-0.8, -0.8]]), np.array([[-grid.TR, grid.ΛF]])),
#         "Apex Vertex": (np.array([[0.0, 1.0]]), np.array([[0.0, grid.ΛC]])),
#         "Base-Right Vertex": (np.array([[0.8, -0.5]]), np.array([[grid.TR, grid.ΛF]])),
#         "Base-Left Vertex": (np.array([[-0.8, -0.5]]), np.array([[-grid.TR, grid.ΛF]])),
#     }
#
#     for name, (input_xy, expected_xy) in test_cases.items():
#         # Ensure the test uses the user's clamp function name
#         x, y, _ = grid.clamp(input_xy[:, 0], input_xy[:, 1], mode=1)
#         result_xy = np.array([x, y]).T
#         # result_xy = grid.clamp(input_xy.copy(), mode=1)
#         assert_allclose(result_xy, expected_xy, atol=1e-5, err_msg=f"Case '{name}' failed")
#
#
# def test_clamp_down_mode(const_grid):
#     """Tests the clamping logic for the DOWN (V) triangle."""
#     grid = const_grid
#     test_cases = {
#         "Inside": (np.array([[-0.1, -0.1]]), np.array([[-0.1, -0.1]])),
#         "Pole": (
#             np.array([[0.000000000000001, -0.8164965132120238]]), np.array([[0.000000000000001, -0.8164965132120238]])),
#         "Outside Apex": (np.array([[0.1, -1.0]]), np.array([[0.0, grid.VF]])),
#         "Outside Left Slant": (np.array([[-0.8, 0.2]]), np.array([[-0.586875, 0.2]])),
#         "Outside Base-Right": (np.array([[0.8, 0.8]]), np.array([[grid.TR, grid.VC]])),
#         "Apex Vertex": (np.array([[0.1, -1.0]]), np.array([[0.0, grid.VF]])),
#         "Base-Right Vertex": (np.array([[0.8, 0.8]]), np.array([[grid.TR, grid.VC]])),
#         "Base-Left Vertex": (np.array([[-0.8, 0.8]]), np.array([[-grid.TR, grid.VC]])),
#     }
#
#     for name, (input_xy, expected_xy) in test_cases.items():
#         x, y, _ = grid.clamp(input_xy[:, 0], input_xy[:, 1], mode=0)
#         result_xy = np.array([x, y]).T
#         assert_allclose(result_xy, expected_xy, atol=1e-5, err_msg=f"Case '{name}' failed")


def test_pc_c1_reverse_map(reg_grid: GridRegions):
    pc = reg_grid.pc_c1()        # (num_regions, num_regions) -> c1
    ugc = reg_grid.ugc_lut       # (num_regions, props), only mode here
    chd = reg_grid.child_lut()   # (2,3,3)

    for par in reg_grid.in_regions:
        mode = ugc[par, reg_grid.mode]
        # each c1 -> its 3 children must map back to c1
        for c1 in (0, 1, 2):
            children = chd[mode, c1]
            for child in children:
                assert pc[par, child] == c1, f"pc_c1 wrong: par=0x{par:02x}, child=0x{child:02x}"

        # optional: children from the *other* mode should not collide to the same c1
        other_mode = 1 - mode
        for other_c1 in (0, 1, 2):
            for alien in chd[other_mode, other_c1]:
                # We can’t assert a specific value (it’s undefined), but we *can* assert it’s not
                # equal to all three c1 values at once; the safe minimal check is that it’s not
                # erroneously identical to the first one we tested above:
                pass  # intentionally no strict assertion — undefined outside parent’s mode


# def test_terminate_normalises_tail(reg_grid: GridRegions):
#     pc = reg_grid.pc_c1()
#     chd = reg_grid.child_lut()
#     ugc = reg_grid.ugc_lut
#
#     # Build a tiny address set with known last/child
#     par = reg_grid.in_regions[0]
#     mode = ugc[par, reg_grid.mode]
#     c1 = 1
#     child = chd[mode, c1, 0]           # not the mode-locked one
#     want = chd[mode, c1, 2]            # mode-locked child
#
#     addr = np.full((1, 5), reg_grid.invalid_ugc, dtype=np.uint8)
#     addr[0, -2] = par
#     addr[0, -1] = child
#
#     out = reg_grid.terminate(addr.copy())
#     assert out[0, -1] == want
#
#
# def test_terminate_keeps_invalid_pairs(reg_grid: GridRegions):
#     inv = reg_grid.invalid_ugc
#     # invalid last and/or child: should remain untouched
#     addr = np.array([[10, inv, inv, inv, inv]], dtype=np.uint8)
#     out = reg_grid.terminate(addr.copy())
#     assert np.array_equal(out, addr)
#

def test_encoder_tail_is_mode_locked(reg_grid: GridRegions):
    # Pick a few random interior points/modes
    rng = np.random.default_rng(42)
    N = 8
    x = rng.uniform(-0.3, 0.3, size=N)
    y = rng.uniform(-0.3, 0.3, size=N)
    mode = rng.integers(0, 2, size=N, dtype=np.uint8)
    addr = reg_grid.ugc_regions(np.array([x, y]).T, mode, depth=18)
    out = reg_grid.terminate(addr.copy())
    assert np.array_equal(out, addr), "terminate should be idempotent on encoder output"


def test_rc_regions(const_grid):
    """Test known input cases to ensure en_grid produces expected region IDs."""
    test_cases = [
        # format: ((ẋ, y), expected_hpn) - needs populating!
        # Core regions
        ((0.0, 0.0), (3, 2, 2)),  # Center point
        ((0.0, 0.5), (1, 1, 2)),  # Top-middle
        ((0.0, -0.5), (4, 2, 1)),  # Bottom-middle
        ((0.5, 0.0), (3, 2, 2)),  # Middle-right
        ((-0.5, 0.0), (3, 1, 1)),  # Middle-left

        # Extreme corner regions
        ((0.0, 1.0), (0, 0, 3)),  # Far top
        ((0.0, -1.0), (5, 3, 0)),  # Far bottom
        ((1.0, 0.0), (3, 3, 3)),  # Far right
        ((-1.0, 0.0), (3, 0, 0)),  # Far left

        # Regions defined by slanted boundaries
        ((0.5, 0.5), (1, 2, 3)),  # Top-right
        ((-0.5, -0.5), (4, 2, 0)),  # Bottom-left
    ]
    for (ẋ_val, y_val), (h, p, n) in test_cases:
        result = const_grid.region_classification(np.array([ẋ_val]), np.array([y_val]))[0]
        expected = (h << 4) | (p << 2) | n
        assert result == expected, f"For ({ẋ_val}, {y_val}) expected {expected}, got {result}"


def test_rc_boundaries(const_grid):
    """Tests the precise vertices and seam midpoints of the grid."""
    # Use f-strings to create descriptive test case names
    test_cases = {
        # --- UP Triangle (Λ) Vertices ---
        "UP Apex": ((0, const_grid.ΛC), (1, 1, 2)),
        "UP Base-Right": ((const_grid.H, const_grid.ΛF), (4, 3, 3)),
        "UP Base-Left": ((-const_grid.H, const_grid.ΛF), (4, 0, 0)),

        # --- DOWN Triangle (V) Vertices ---
        "DOWN Apex": ((0, const_grid.VF), (4, 2, 1)),
        "DOWN Base-Right": ((const_grid.H, const_grid.VC), (2, 3, 3)),
        "DOWN Base-Left": ((-const_grid.H, const_grid.VC), (2, 0, 0)),

        # --- Seam Midpoints ---
        "UP Base Midpoint": ((0, const_grid.ΛF), (4, 2, 1)),
        "DOWN Base Midpoint": ((0, const_grid.VC), (2, 1, 2)),
        "Right Vertical Seam": ((const_grid.TR * const_grid.R3, 0), (3, 3, 3)),  # Same as (0.5, 0) approx.

        # --- Origin ---
        "Origin": ((0.0, 0.0), (3, 2, 2)),
    }

    for name, ((ẋ_val, y_val), (h, p, n)) in test_cases.items():
        ẋ_arr = np.array([ẋ_val])
        y_arr = np.array([y_val])

        result = const_grid.region_classification(ẋ_arr, y_arr)[0]
        expected = (h << 4) | (p << 2) | n

        assert result == expected, f"Case '{name}': For ({ẋ_val}, {y_val}) expected {expected}, got {result}"


def test_rc_internal_seams(const_grid):
    """Tests points on the internal grid lines (y=0, y=±ẋ)."""
    test_cases = {
        # --- Points on the y=0 axis ---
        "Positive X-Axis": ((0.1, 0.0), (3, 2, 2)),
        "Negative X-Axis": ((-0.1, 0.0), (3, 1, 1)),

        # --- Points on the y = ẋ line ---
        "y=x in Quadrant 1": ((0.2, 0.2), (2, 2, 2)),
        "y=x in Quadrant 3": ((-0.2, -0.2), (3, 2, 1)),

        # --- Points on the y = -ẋ line ---
        "y=-x in Quadrant 2": ((-0.2, 0.2), (2, 1, 2)),
        "y=-x in Quadrant 4": ((0.2, -0.2), (3, 2, 2)),

    }

    for name, ((ẋ_val, y_val), (h, p, n)) in test_cases.items():
        ẋ_arr = np.array([ẋ_val])
        y_arr = np.array([y_val])

        result = const_grid.region_classification(ẋ_arr, y_arr)[0]
        expected = (h << 4) | (p << 2) | n

        assert result == expected, f"Case '{name}': For ({ẋ_val}, {y_val}) expected {expected}, got {result}"


def test_rc_batch_processing(const_grid):
    """Tests the function with a batch of multiple points at once."""
    # Combine several known cases into batch arrays
    ẋ_batch = np.array([0.0, 0.5, -0.2])
    y_batch = np.array([0.0, 0.5, -0.2])

    # Calculate the expected results for the batch
    expected_h = np.array([3, 1, 3])
    expected_p = np.array([2, 2, 2])
    expected_n = np.array([2, 3, 1])
    expected_results = (expected_h << 4) | (expected_p << 2) | expected_n

    # Run the classification on the entire batch
    batch_results = const_grid.region_classification(ẋ_batch, y_batch)

    # Use NumPy's testing utility to compare the arrays
    np.testing.assert_array_equal(batch_results, expected_results)


def test_rc_default_case(const_grid):
    """Tests a point far outside the grid to check default case handling."""
    # This point is far to the right and bottom
    ẋ_val, y_val = 2.0, -2.0

    # Manually determine the expected default IDs
    # y < VF -> h_id should be default=5
    # y-ẋ < -Ẇ -> p_id should be default=3
    # y+ẋ = 0 -> n_id should be 2
    h, p, n = 5, 3, 2

    expected = (h << 4) | (p << 2) | n
    result = const_grid.region_classification(np.array([ẋ_val]), np.array([y_val]))[0]

    assert result == expected, f"Default case failed: expected {expected}, got {result}"

def test_encode_one_step_child_is_legal(reg_grid: GridRegions):
    rng = np.random.default_rng(0)
    # points around the origin so we’re well inside; both modes
    x = rng.uniform(-0.1, 0.1, size=16)
    y = rng.uniform(-0.1, 0.1, size=16)
    mode = np.array([0]*8 + [1]*8, dtype=np.uint8)

    addr = reg_grid.ugc_regions(np.column_stack([x, y]), mode, depth=1)
    ugc = reg_grid.ugc_lut
    chd = reg_grid.child_lut()
    pc  = reg_grid.pc_c1()

    # root at [0], first child at [1]
    par = addr[:, 0]
    cur = addr[:, 1]

    for i, (p, c) in enumerate(zip(par, cur)):
        assert c != reg_grid.invalid_ugc, f"row {i}: invalid child at depth=1"
        pmo = ugc[p, reg_grid.mode]
        c1  = pc[p, c]
        assert c1 in (0,1,2), f"row {i}: bad c1=0x{c1:x} for par=0x{p:02x}, cur=0x{c:02x}"
        assert c in chd[pmo, c1], f"row {i}: child 0x{c:02x} not in legal children of par=0x{p:02x} mode={pmo}"

def _manual_one_step(reg: GridRegions, xy, mode):
    """Replicate exactly one encoder step: clamp → classify → subtract offsets → ×3 (classifier plane)."""
    x = xy[:,0].copy()
    y = xy[:,1].copy()

    # clamp per parent mode; get xbar
    x, y, xbar = reg.clamp(x, y, mode)

    # classify
    region = reg.region_classification(xbar, y)
    props  = reg.ugc_lut[region]
    in_scope = np.where(mode == 1, props[:, reg.in_up], props[:, reg.in_dn])
    rid = np.where(in_scope, region, reg.invalid_ugc)

    # step (in classifier plane)
    off_ẋ = reg.ugc_off_ẋ[rid]
    off_y = reg.ugc_off_y[rid]
    xbar2 = (xbar - off_ẋ) * 3.0
    y2    = (y    - off_y) * 3.0
    x2    = xbar2 / reg.R3

    # update mode only for valid rows
    child_mode = reg.ugc_lut[rid, reg.mode]
    valid = (rid != reg.invalid_ugc)
    mode2 = np.where(valid, child_mode, mode).astype(np.uint8)

    return rid, np.column_stack([x2, y2]), mode2

def test_first_step_math_matches_encoder(reg_grid: GridRegions):
    rng = np.random.default_rng(1)
    N = 32
    xy   = rng.uniform(-0.2, 0.2, size=(N,2))
    mode = rng.integers(0,2,size=N, dtype=np.uint8)

    addr = reg_grid.ugc_regions(xy, mode, depth=1)
    rid1 = addr[:,1]

    rid_m, xy_m, mode_m = _manual_one_step(reg_grid, xy, mode)
    np.testing.assert_array_equal(rid_m, rid1, "first child region mismatch")
    # encode again for one more step from both states; they should stay aligned
    addr2 = reg_grid.ugc_regions(xy_m, mode_m, depth=1)
    np.testing.assert_array_equal(addr2[:, 0], addr[:, 0], "root must remain the same after one step")

@pytest.mark.parametrize("depth", [2, 3])
def test_shallow_roundtrip(reg_grid: GridRegions, depth):
    rng = np.random.default_rng(2)
    N = 20
    xy   = rng.uniform(-0.25, 0.25, size=(N,2))
    mode = rng.integers(0,2,size=N, dtype=np.uint8)

    addr = reg_grid.ugc_regions(xy, mode, depth=depth)
    dec  = reg_grid.ugc_dec(addr)
    np.testing.assert_allclose(dec[:,:2], xy, atol=1e-15, err_msg="decode(xy) mismatch (shallow)")
    np.testing.assert_array_equal(dec[:,2].astype(np.uint8), mode, err_msg="decode(mode) mismatch (shallow)")

    re_addr = reg_grid.ugc_regions(dec[:,:2], dec[:,2].astype(np.uint8), depth=depth)
    np.testing.assert_array_equal(re_addr, addr, "encode(decode(addr)) mismatch (shallow)")

def test_boundary_midpoints_do_not_go_invalid_early(reg_grid: GridRegions):
    g = reg_grid
    # a few exact boundary points in classifier plane, mapped back to x
    xbar = np.array([0.0,  g.Ẇ, -g.Ẇ])
    y    = np.array([0.0,  g.VC, g.ΛF])   # pick mid/base lines
    x    = xbar / g.R3

    # test both modes
    XY = np.column_stack([np.tile(x,3), np.repeat(y,3)])
    mode = np.array([0,0,0, 1,1,1, 0,1,0], dtype=np.uint8)[:XY.shape[0]]

    addr = g.ugc_regions(XY, mode, depth=3)
    # none of first 3 children should be invalid
    assert not np.any(addr[:,1:4] == g.invalid_ugc), f"invalid child within first 3 steps on boundary inputs"

def test_terminate_always_normalises_tail_simple(reg_grid: GridRegions):
    pc  = reg_grid.pc_c1()
    chd = reg_grid.child_lut()
    ugc = reg_grid.ugc_lut
    par = reg_grid.in_regions[0]
    mode = ugc[par, reg_grid.mode]
    # choose the non-mode-locked child 0
    child0 = chd[mode, 1, 0]
    want   = chd[mode, 1, 2]

    addr = np.full((1, 6), reg_grid.invalid_ugc, dtype=np.uint8)
    addr[0, -2] = par
    addr[0, -1] = child0
    out = reg_grid.terminate(addr.copy())
    assert out[0, -1] == want

def test_regression_named_points_depth3(reg_grid: GridRegions):
    # NAΛ centroid and the SPΛ “bad girl”
    pts  = np.array([
        [-0.210025776447247, -0.020916439797272],
        [+0.105560608419664, -0.134651734661425],
    ])
    mode = np.array([1, 1], dtype=np.uint8)
    addr = reg_grid.ugc_regions(pts, mode, depth=3)
    dec  = reg_grid.ugc_dec(addr)
    np.testing.assert_allclose(dec[:,:2], pts, atol=1e-15)
    np.testing.assert_array_equal(dec[:,2].astype(np.uint8), mode)

def test_lateral_symmetry(reg_grid: GridRegions):
    ugc = reg_grid.ugc_lut
    chd = reg_grid.child_lut()
    bad = []
    for par in reg_grid.in_regions:
        mo = ugc[par, reg_grid.mode]
        c0, _, _ = chd[mo, 0]   # lateral A, center, lateral B
        c2, _, _ = chd[mo, 2]
        y0 = reg_grid.ugc_off_y[c0]
        y2 = reg_grid.ugc_off_y[c2]
        x0 = reg_grid.ugc_off_ẋ[c0]
        x2 = reg_grid.ugc_off_ẋ[c2]
        if not np.isclose(y0, y2, atol=1e-15) or not np.isclose(x0, -x2, atol=1e-15):
            bad.append((par, (x0, y0), (x2, y2)))
    assert not bad, "lateral offsets not symmetric: " + ", ".join(f"par={hex(p)} L={a} R={b}" for p,a,b in bad)

def test_offsets_self_classify(reg_grid: GridRegions):
    bad = []
    for rid in reg_grid.in_regions:
        rid2 = reg_grid.region_classification(
            np.array([reg_grid.ugc_off_ẋ[rid]]),
            np.array([reg_grid.ugc_off_y[rid]])
        )[0]
        if rid2 != rid:
            bad.append((rid, rid2))
    assert not bad, f"offsets classify to different regions: {[(hex(a), hex(b)) for a,b in bad]}"

def test_reg_valid_points(reg_grid):
    """
    Tests the addresses generation for a known coordinate that should succeed.
    """
    test_cases = np.array([
        [-0.210025776447247, -0.020916439797272, 1],  # NAΛ: centroid (easy)
        [+0.105560608419664, -0.134651734661425, 1],  # SPΛ: bad girl
    ])
    names = ['NAΛ: centroid', 'SPΛ: bad girl']
    for name, test_case in zip(names, test_cases):
        xy = np.array([test_case[:2]])
        mode = np.array([test_case[2]])
        result_address = reg_grid.ugc_regions(xy, mode)

        # New: structural/local correctness for first k steps
        k = 33
        addr = result_address[0, :k]
        # 1) root must be consistent with initial mode
        root_expected = 0x16 if mode[0] == 1 else 0x49
        assert addr[0] == root_expected

        # 2) each step must be a valid child of its parent in the parent’s mode
        ugc = reg_grid.ugc_lut  # (num_regions, 1) -> mode
        chd = reg_grid.child_lut()  # (2,3,3)
        c1r = reg_grid.pc_c1()  # (num_regions, num_regions) -> c1

        for i in range(1, k):
            par = addr[i - 1]
            cur = addr[i]
            par_mode = ugc[par, reg_grid.mode]
            # if either par or cur is invalid, we bail to let roundtrip catch it
            if par == reg_grid.invalid_ugc or cur == reg_grid.invalid_ugc:
                raise AssertionError(f"Invalid region seen at step {i}: par={par:#04x}, cur={cur:#04x}")
            c1 = c1r[par, cur]
            # c1 must be 0/1/2 and cur must be one of the canonical 3 children
            assert c1 in (0, 1, 2), f"{name}: Bad c1 at step {i}: c1={c1}, par={par:#04x}, cur={cur:#04x}"
            assert cur in chd[par_mode, c1], f"{name}: disallowed child; step {i}: par={par:#04x}, cur={cur:#04x}, mode={par_mode}"

        # Strong invariant: decode → re-encode at same depth should be identical
        depth = result_address.shape[1] - 2
        dec = reg_grid.ugc_dec(result_address)
        assert_array_equal(dec[:, :2], xy, err_msg=f'{name}: Roundtrip address mismatch.')
        assert_array_equal(dec[:, 2], mode, err_msg=f'{name}: Roundtrip mode mismatch.')
        re_addr = reg_grid.ugc_regions(dec[:, :2], dec[:, 2].astype(np.uint8), depth=depth)
        assert_array_equal(re_addr, result_address, err_msg=f'{name}: Roundtrip mismatch.')


def test_ugc_regions_invalid_point(reg_grid):
    """
    Tests the addresses generation for a coordinate that is logs of bounds.
    """
    # Inputs for a point far outside the grid
    x = np.array([10.0])
    y = np.array([10.0])
    mode = np.array([1])  # UP mode

    # Generate the addresses
    result_address = reg_grid.ugc_regions(np.array([x, y]).T, mode, depth=5)

    # The second region (index 1) should be marked as invalid because the
    assert result_address[0, 1] == reg_grid.invalid_ugc


def test_roundtrip_conversion(reg_grid):
    """
    Tests the full encode-decode roundtrip process.
    """
    # 1. Arrange: Define the initial test point and mode.
    initial_x = np.array([0.278558759260123456789])
    initial_y = np.array([0.293862554740123456789])
    initial_mo = np.array([0])  # DOWN mode
    initial_xym = np.stack([initial_x, initial_y, initial_mo], axis=-1)

    # 2. Act: Run the full encode and decode cycle.
    # Encode the coordinate to a URI addresses.
    xy = np.array([initial_x, initial_y]).T
    uri_address = reg_grid.ugc_regions(xy, initial_mo)

    # Decode the URI addresses back to a coordinate.
    decoded_xym = reg_grid.ugc_dec(uri_address)
    # display(initial_xym, decoded_xym)

    # 3. Assert: Check if the result is close to the original.
    assert_allclose(decoded_xym, initial_xym, atol=1e-40)


def test_ext_neighbours(rel_grid):
    """
    Neighbours
    """
    # Rapa Nui Moai
    reg01 = np.array([[0x49, 0x35, 0x25, 0x3a, 0x21, 0x2b, 0x49, 0x25, 0x26, 0x35, 0x26,
                       0x25, 0x16, 0x3e, 0x34, 0x34, 0x39, 0x3a, 0x21, 0x2b, 0x2b]])
    ref01 = np.array([[0x49, 0x35, 0x25, 0x3a, 0x21, 0x2b, 0x49, 0x25, 0x26, 0x35, 0x26,
                       0x25, 0x16, 0x3e, 0x34, 0x34, 0x39, 0x2a, 0x34, 0x3e, 0x3e]])
    # North Pole
    reg02 = np.array([[0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49,
                       0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49,
                       0x49, 0x49, 0x49, 0x49, 0x49, 0x49]])
    ref02 = np.array([[0x16, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34,
                       0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34, 0x34,
                       0x34, 0x34, 0x34, 0x34, 0x34, 0x34]])

    rnm = rel_grid.region_neighbours(reg01)
    assert_array_equal(rnm, ref01)
    npo = rel_grid.region_neighbours(reg02)
    assert_array_equal(npo, ref02)


def test_neighbour_decode_encode_roundtrip_reg01(rel_grid, reg_grid):
    """Neighbour → decode(xy,mode) → re-encode should roundtrip to the same neighbour addresses.
    Exercises the transform-then-classify pipeline (no manual C1 flipping), including cascades.
    """
    # Rapa Nui Moai case from test_ext_neighbours
    reg01 = np.array([[
        0x49, 0x35, 0x25, 0x3a, 0x21, 0x2b, 0x49, 0x25, 0x26, 0x35, 0x26,
        0x25, 0x16, 0x3e, 0x34, 0x34, 0x39, 0x3a, 0x21, 0x2b, 0x2b
    ]])

    # 1) Neighbour addresses via geometric LUT (with cascading where needed)
    nb_addr = rel_grid.region_neighbours(reg01)

    # 2) Decode neighbour addresses back to (x, y, initial_mode)
    xym = reg_grid.ugc_dec(nb_addr)
    mode = xym[:, 2].astype(np.uint8)

    # 3) Re-encode using decoded coordinates + initial mode, for the SAME depth
    depth = nb_addr.shape[1] - 2

    re_addr = reg_grid.ugc_regions(xym[:, :2], mode, depth=depth)

    # 4) Should match exactly
    assert_array_equal(re_addr, nb_addr)


def test_neighbour_decode_encode_roundtrip_reg02(rel_grid, reg_grid):
    """Same roundtrip for the Polar (seam-heavy) case.
    Confirms seam hops are handled by geometry + reclassification alone.
    """
    # North Pole case from test_ext_neighbours
    reg02 = np.array([[
        0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49,
        0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49,
        0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49
    ]])

    nb_addr = rel_grid.region_neighbours(reg02)

    xym = reg_grid.ugc_dec(nb_addr)
    mode = xym[:, 2].astype(np.uint8)

    depth = nb_addr.shape[1] - 2
    xy = np.stack([xym[:, 0], xym[:, 1]], axis=-1)
    re_addr = reg_grid.ugc_regions(xy, mode, depth=depth)
    assert_array_equal(re_addr, nb_addr)


def test_trace_roundtrip(reg_grid: GridRegions):
    xy = np.array([[-0.210025776447247, -0.020916439797272]])
    mode = np.array([1], dtype=np.uint8)
    reg_grid.trace_roundtrip(xy, mode, depth=8, name="NAΛ centroid")

def test_diag_first_step(reg_grid: GridRegions):
    xy = np.array([[-0.210025776447247, -0.020916439797272]])
    mode = np.array([1], dtype=np.uint8)
    reg_grid.diag_first_step(xy, mode, depth=8, name="NAΛ centroid")




# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
AK Octahedral Projection 'oct_ell'
"""
from functools import cache
import numpy as np
from numpy.typing import NDArray
from hhg9 import Points
from hhg9.base.projection import Projection
from hhg9.algorithms.distance import haversine_rad
from hhg9.algorithms import find_coords
from pyproj import CRS
from hhg9.h9 import H9C, H9K


class AKOctahedralEllipsoid(Projection):
    """
        An Octahedron/Ellipsoid Projection generated via an analytical approximation to a
        force-directed dataset. Approximation designer: Anders Kaseorg
    """
    ALPHA = 3.227806237143884260376580  # 𝛂 - vis. Kaseorg.

    def __init__(self, registrar, name='oct_ell'):
        self.reg = registrar
        super().__init__(self.reg, name, 'c_oct', 'c_ell')
        crs_ecef = CRS.from_epsg(4978)  # WGS84 ECEF (x, y, z)
        ecef_e = crs_ecef.ellipsoid
        self.ab = ecef_e.semi_major_metre, ecef_e.semi_minor_metre
        self.ab2 = 1., (self.ab[1] / self.ab[0]) ** 2

        self.b_oct = self.reg.domain('b_oct')
        self.c_oct = self.reg.domain('c_oct')
        self.c_ell = self.reg.domain('c_ell')
        self.g_gcd = self.reg.domain('g_gcd')
        self.vertices = np.array(list(self.rev_cs.vertices.values()))
        self._e = 1e-200
        self.tol = 1e-15

        # Level 0 area ≈ 5362 km (area of Earth / 12 hexes)
        earth = 510_065_621_724_154.6
        self.hex_0 = earth / 12
        self.tri_0 = earth / 8
        l_hex = self.hex_0
        l_tri = self.tri_0
        self.h_areas = np.zeros((64,), dtype=np.float64)
        self.t_areas = np.zeros((64,), dtype=np.float64)
        for i in range(64):
            self.h_areas[i] = l_hex
            self.t_areas[i] = l_tri
            l_hex /= 9
            l_tri /= 9
        self.accuracy = 34  # accuracy is nanometres.

    @cache
    def rad_gcd(self):
        """Return the radians GCD domain, registering it (and its deg⟷rad projection) on first use.
        This is lazy and idempotent: repeated calls reuse the same registrar entries.
        """
        # Imports kept local to avoid import cycles at module import time
        from hhg9.domains import RadiansGCD
        from hhg9.projections import RGCD_GCD

        reg = self.reg
        try:
            return reg.domain('r_gcd')
        except Exception:
            # not yet registered; fall through to create & wire projections
            pass

        r_gcd = RadiansGCD(reg)  # registers the domain
        RGCD_GCD(reg)  # registers deg↔rad projections (idempotent if already present)
        return r_gcd

    def _invariants(self, v):
        """Return invariant points: Those which are on the vertices themselves"""
        diff = np.abs(v[..., None, :] - self.vertices[..., :, :])  # shape (?, 8, 3)
        matches = np.all(diff < self.tol, axis=-1)  # shape (1000, 8)
        return np.array(np.any(matches, axis=-1))  # indices of v

    def _core_raw(self, uvw: NDArray) -> NDArray:
        """
        Vectorized core projection: maps points from the unit octahedron to the unit sphere.
        Handles edge cases where one coordinate is near zero (i.e., edge of the octant).
        Returns the un-normalised projection vector.
        """
        uvw = np.asarray(uvw)
        α = self.ALPHA
        e = self._e

        t_uvw = np.tan((np.pi * uvw + e) * 0.5)
        xu, xv, xw = t_uvw[..., -3], t_uvw[..., -2], t_uvw[..., -1]
        u2, v2, w2 = xu ** 2, xv ** 2, xw ** 2

        # Default calculation
        y0p = np.asarray(xu * (v2 + w2 + α * w2 * v2) ** 0.25)
        y1p = np.asarray(xv * (u2 + w2 + α * u2 * w2) ** 0.25)
        y2p = np.asarray(xw * (u2 + v2 + α * u2 * v2) ** 0.25)
        pv = np.stack([y0p, y1p, y2p], axis=-1)
        return pv

    def _core(self, uvw: NDArray) -> NDArray:
        pv = self._core_raw(uvw)
        return self.normalise(pv)

    def get_accuracy(self, layer):
        """
        Get accuracy in m2 from a given level.
        """
        if layer < 0:
            raise ValueError("get_accuracy: layer must be >= 0")
        h_area = self.h_areas[-1]
        t_area = self.t_areas[-1]
        if layer < len(self.h_areas):
            h_area = self.h_areas[layer]
            t_area = self.t_areas[layer]
        # side = np.sqrt((2*area)/(3*H9K.R3))
        return h_area, t_area

    def set_accuracy(self, m2):
        """
        Set the level such that the hex area is ≤ desired accuracy in m2.
        """
        idx = np.searchsorted(self.h_areas[::-1], m2, side='right')
        self.accuracy = len(self.h_areas) - idx
        return self.accuracy

    def normalise(self, p):
        """Normalise result to elliptical coordinates"""
        xx, yy, zz = p[..., 0], p[..., 1], p[..., 2]
        a2, b2 = self.ab2
        n = np.sqrt((xx ** 2 + yy ** 2) / a2 + zz ** 2 / b2)
        return np.stack([xx / n, yy / n, zz / n], axis=-1)

    def forward(self, arr: Points) -> Points:
        """
        c_oct->c_ell projection
        Convert a NDArray of octahedral points projected onto WGS84 Ellipsoid
        Anders Kaseorg: https://math.stackexchange.com/questions/5016695/
        :param pts:  An array of Euclidean points on the surface of a unit octahedron.
        :return: Points UVW on WGS84 Ellipsoid
        """
        xyz = arr.coords
        sgn = np.sign(xyz)
        # Run core forward on a copy so we don't lose the original signs
        uvw = np.asarray(xyz, dtype=np.float64).copy()  # avoid mutating caller
        aa = self._invariants(uvw)
        if np.any(~aa):
            trx = self._core(uvw[~aa])
            uvw[~aa] = trx
        if np.any(aa):
            # map them to the ellipsoid as well
            uvw[aa] = self.normalise(uvw[aa])
        core_abs = np.abs(uvw)
        # Apply signs: components with sign==0 remain exactly 0
        res = self.ab[0] * (core_abs * sgn)
        return Points(res, domain=self.fwd_cs, samples=arr.samples, components=arr.components)

    def backward(self, arr: Points) -> Points:
        """
        c_ell->c_oct projection using find_coords rootfinding algorithm.
        """
        r_gcd = self.rad_gcd()
        if arr.components is None:
            self.rev_cs.binning(arr)  # We need the octant identity for each point.
        uvw = arr.copy()
        cmp = uvw.components
        # cmp = uvw.components[:, np.newaxis, :]  # use this for referring to the points' octant identity.
        rll = self.reg.project(uvw, [self.c_ell, self.g_gcd, r_gcd])  # Project to give us GCD reference values.
        ref = rll.coords  # reference addresses
        _, oct_m = uvw.cm()  # we want their modes.

        def fwd(xy, octants):
            """Project contender xy (in barycentric) to GCD"""
            coords = Points(xy.reshape(-1, 2), self.b_oct, octants.reshape(-1, 3))
            grx = self.reg.project(coords, [self.b_oct, self.c_oct, self.c_ell, self.g_gcd, r_gcd])
            return grx.coords.reshape(xy.shape)

        found, _ = find_coords(ref, oct_m, cmp, H9C, fwd, haversine_rad, self.accuracy+2, beam_width=6)
        bpt = Points(found, self.b_oct, uvw.components)
        return self.reg.project(bpt, [self.b_oct, self.rev_cs])  # rev_cs = c_oct

    def _core_jacobian(self, uvw):
        """
        :param uvw:
        :return:
        """
        alpha = self.ALPHA
        # match the epsilon shift used in _core to avoid tiny inconsistencies
        t = np.tan((np.pi * uvw + self._e) * 0.5)
        s = (np.pi * 0.5) * (1 + t * t)  # d/du tan(πu/2)

        u, v, w = t[..., 0], t[..., 1], t[..., 2]
        u2, v2, w2 = u * u, v * v, w * w

        def f_pow(b, c):
            # (b + c + alpha*b*c) ** 1/4
            return (b + c + alpha * b * c) ** 0.25

        def df_db(b, c):
            return 0.25 * (b + c + alpha * b * c) ** (-0.75) * (1 + alpha * c)

        def df_dc(b, c):
            return 0.25 * (b + c + alpha * b * c) ** (-0.75) * (1 + alpha * b)

        g0 = f_pow(v2, w2)
        g1 = f_pow(u2, w2)
        g2 = f_pow(u2, v2)

        # partials for y0, y1, y2 w.r.t. (u, v, w)
        dy0_du = s[..., 0] * g0
        dy0_dv = u * s[..., 1] * df_db(v2, w2) * 2 * v
        dy0_dw = u * s[..., 2] * df_dc(v2, w2) * 2 * w

        dy1_du = v * s[..., 0] * df_db(u2, w2) * 2 * u
        dy1_dv = s[..., 1] * g1
        dy1_dw = v * s[..., 2] * df_dc(u2, w2) * 2 * w

        dy2_du = w * s[..., 0] * df_db(u2, v2) * 2 * u
        dy2_dv = w * s[..., 1] * df_dc(u2, v2) * 2 * v
        dy2_dw = s[..., 2] * g2

        j = np.empty(uvw.shape[:-1] + (3, 3), dtype=uvw.dtype)
        j[..., 0, 0] = dy0_du
        j[..., 0, 1] = dy0_dv
        j[..., 0, 2] = dy0_dw
        j[..., 1, 0] = dy1_du
        j[..., 1, 1] = dy1_dv
        j[..., 1, 2] = dy1_dw
        j[..., 2, 0] = dy2_du
        j[..., 2, 1] = dy2_dv
        j[..., 2, 2] = dy2_dw
        return j

    def _norm_jacobian(self, p):
        a2, b2 = self.ab2
        x, y, z = p[..., 0], p[..., 1], p[..., 2]
        n = np.sqrt((x ** 2 + y ** 2) / a2 + z ** 2 / b2)

        # Mp (ellipsoidal metric times p)
        mp = np.stack([x / a2, y / a2, z / b2], axis=-1)  # (...,3)

        eye = np.eye(3)
        eye = np.broadcast_to(eye, p.shape[:-1] + (3, 3))  # (...,3,3)

        # make n broadcastable
        n2 = n[..., None, None]
        mp_scaled = mp / (n[..., None] ** 2)

        # Df = (I - p ⊗ (Mp)/n**2) / n
        outer = p[..., :, None] * mp_scaled[..., None, :]  # (...,3,3)
        return (eye - outer) / n2

    def jacobian(self, uvw):
        """
        :param uvw:
        :return: jacobian of the forward projection
        """
        uvw = np.asarray(uvw, dtype=np.float64, copy=True)
        aa = self._invariants(uvw)
        # Invariant points: match the forward() behaviour x ↦ a * x
        # i.e. take J ≈ a * I (this is a convention, but consistent with the code path)
        if np.any(aa):
            a = uvw[aa]
            a[a == 0] = np.finfo(np.float64).tiny
            c = np.sign(a)/np.sqrt(3)
            adj = a + 1e-15 * (c - a)
            uvw[aa] = adj

        # Non-invariant points: current analytic Jacobian
        # if np.any(~aa):
        u = uvw
        y_raw = self._core_raw(u)
        y = self.normalise(y_raw)
        j_core = self._core_jacobian(u)
        j_norm = self._norm_jacobian(y_raw)
        j_sub = j_norm @ j_core

        sgn = np.sign(u)
        dabs = np.sign(y) * sgn
        j_sub = dabs[..., :, None] * j_sub

        j = self.ab[0] * j_sub

        return j





# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
GeneralGCD is recognised as [latitude,longitude]
"""

import numpy as np
from hhg9.base import Points
from hhg9.base.domain import Domain
from numpy.typing import NDArray


class GeneralGCD(Domain):
    """Latitude and Longitude GCD"""

    def __init__(self, registrar):
        super().__init__(registrar, 'g_gcd', 2)
        self.lat_min, self.lat_max = (-90., 90.)
        self.lon_min, self.lon_max = (-180., 180.)

    def valid(self, pts: Points) -> NDArray:
        lat, lon = pts.coords[..., 0], pts.coords[..., 1]
        return np.logical_and(
            np.abs(lat) <= 90,
            np.abs(lon) <= 180
        )

    def adopt(self, pts: NDArray):
        """
        Take an array and adopt as this domain.
        """
        if pts.ndim == 2 and pts.shape[1] == 2:
            return Points(pts, self)
        raise ValueError(f'{pts.shape} seems wrong for latitude/longitude array.')



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
This is 'c_oct' cartesian octahedral surface xyz
"""

import numpy as np
from numpy.typing import NDArray
from hhg9.base.composite import CompositeDomain, ComponentDomain
from hhg9.base.point_format import PointFormat


class OctantCartesian(ComponentDomain):
    """
    This a 3D side of an Octahedron.
    """

    def __init__(self, registrar, dom, name: str, sign: tuple):
        super().__init__(registrar, name, dom, None, sign, 3)
        self.points = None

    def valid(self, pts: NDArray) -> NDArray:
        """
        Return a boolean mask indicating which 3D points belong to this octant face,
        i.e. their coordinate-wise signs match the face sign triple.
        Accepts a single point shape (3,) or an array of points shape (N,3).
        """
        vals = np.atleast_2d(np.asarray(pts, dtype=float))
        target = np.array(self.sig(), dtype=float)  # (+/-1, +/-1, +/-1)
        matches = np.sign(vals) == target
        mask = np.all(matches, axis=-1)
        return mask if np.ndim(pts) > 1 else mask[0]


class OctahedralCartesian(CompositeDomain):
    """
    Basic octahedral properties and methods in 3D cartesian space.
    self.signs maps sign triple → face name.
    """

    def __init__(self, registrar: object) -> object:
        super().__init__(registrar, 'c_oct', 3)

        # Define properties using symmetry
        self.sides = {}
        self.vertices = {}
        self.signs = {}

        # Stable ids for faces: maintain both name↔id and sign↔id maps
        self.face_id = {}         # face name -> id (0..7)
        self.id_face = {}         # id -> face name
        self.sign_to_id = {}      # sign triple -> id

        # Define axes of symmetry for octahedron
        axes = [['A', 'P'], ['E', 'W'], ['N', 'S']]
        for i, axis in enumerate(axes):
            for j, v in zip([1, -1], axis):
                vertex = [0, 0, 0]
                vertex[i] = j
                self.vertices[v] = tuple(vertex)

        # Generate all face permutations using np.meshgrid
        face_permutations = np.stack(np.meshgrid(*axes), axis=-1).reshape((-1, 3))
        # Generate a canonical list of edges also. Why?
        # It could generate the ID and name of the 12 root0 hexagons.
        self.edges = np.unique(face_permutations[:, [[0, 1], [1, 2], [0, 2]]].reshape((-1, 2)), axis=0)[:, ::-1]

        # self.verts = np.array((8, 3), dtype=np.float64)
        # Define face sides and their properties
        from hhg9 import Points
        for face_arr in face_permutations:
            face = ''.join(face_arr[::-1])  # Reverse order to match face naming
            triple = np.asarray([self.vertices[s] for s in face_arr])
            sign = tuple(np.sum(triple, axis=1).tolist())
            sig = f'{self.name}:{face}'
            self.sides[face] = OctantCartesian(registrar, self, sig, sign)
            self.sides[face].points = triple
            self.signs[sign] = face  # Store based on sum signature
            self.components[sign] = self.sides[face]
        octants = np.array(list(self.components.keys()), dtype=int)
        oids = Points.calc_octant_ids(octants)
        for fid_, triple in zip(oids, octants):
            fid = int(fid_)
            sign = tuple(triple.tolist())
            face = self.signs[tuple(sign)]
            self.face_id[face] = fid
            self.id_face[fid] = face
            self.sign_to_id[tuple(sign)] = fid
            # DO NOT uncomment the below until ready to migrate components entirely to ID.
            # self.components[fid] = self.sides[face]

    @classmethod
    def valid(cls, pts: NDArray) -> NDArray:
        """
        Return an array of bools according to the validity criterion
        :param pts: set of 3d Euclidean points
        """
        pbs = np.sum(np.abs(pts), axis=-1) - 1.
        return np.abs(pbs) < 1e-15

    def register_format(self, af: PointFormat):
        """Decorator to register an AddressFormat for each component."""
        super().register_format(af)
        for side in self.sides:
            self.sides[side].register_format(af)

    def face_for_sign(self, sign: tuple) -> str:
        return self.signs[sign]

    def id_for_sign(self, sign: tuple) -> int:
        return self.sign_to_id[sign]


# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
H9 Polygon Generation.

This module generates the geometric shapes (polygons) associated with H9 grid cells.
It uses Look-Up Tables (LUTs) to define the vertices of triangles, half-hexagons,
and full hexagons relative to a cell center.

**Key Features:**

* **Shape LUTs:** Pre-calculated vertex offsets for all orientations (Mode 0/1, C2 0-2).
* **Mesh Generation:** Creating triangle meshes (`tri_mesh`) for surface plots.
* **Hex Binning:** Aggregating points into specific resolution layers (`hex_layer`).
* **Boundary Handling:** Complex logic to "stitch" hexagons that straddle the edges of the octahedron faces.


**Coordinate System:**
Shapes are defined using offsets of :math:`(U, 3V)` relative to the cell center.
All polygons are defined in **Clockwise** order.
"""

from dataclasses import dataclass
import numpy as np
from numpy.typing import NDArray
from typing import Tuple, List, Optional, Callable, Protocol

from hhg9.h9 import H9C, H9K
from hhg9.h9.addressing import reg_hex_digits, hex_digits_reg, HEX_LUTS, TailStyle, hex_layer, tail_unpack_reversible, \
    hex_key
from hhg9.h9.classifier import location
from hhg9.h9.protocols import H9ConstLike, H9PolygonLike


@dataclass(frozen=True, slots=True)
class H9Polygon:
    """
    Immutable container for Polygon LUTs.

    Attributes:
        hh (NDArray[float64]): Half-hex vertices. Shape (2, 3, 4, 2) -> [Mode, C2, Vertices, XY].
        hx (NDArray[float64]): Full-hex vertices. Shape (2, 3, 6, 2).
        tx (NDArray[float64]): Cell triangle vertices. Shape (2, 3, 3, 3, 2).
        se (NDArray[float64]): Supercell edge points. Shape (2, 9, 2).
        sv (NDArray[float64]): Supercell vertices. Shape (2, 3, 2).
        gd (NDArray[float64]): Unshared points of a cell excluding (0,0).
    """
    hh: NDArray[np.float64]
    hx: NDArray[np.float64]
    tx: NDArray[np.float64]
    se: NDArray[np.float64]
    sv: NDArray[np.float64]
    gd: NDArray[np.float64]


def _h9_polygon(h9k: Optional[H9ConstLike] = None) -> H9Polygon:
    """
    Factory to build the H9Polygon singleton.

    Defines the relative vertices for all standard grid shapes (clockwise order).
    """
    pts = {
        # Clockwise.
        (0, 0): [  # c2 half-hexagons mode 0
            [(3, 1), (2, 0), (0, 0), (-1, 1)],
            [(0, -2), (-1, -1), (0, 0), (2, 0)],
            [(-3, 1), (-1, 1), (0, 0), (-1, -1)]
        ],
        (0, 1): [  # c2 half-hexagons mode 1
            [(-1, -1), (0, 0), (2, 0), (3, -1)],
            [(-1, 1), (0, 0), (-1, -1), (-3, -1)],
            [(2, 0), (0, 0), (-1, 1), (0, 2)]
        ],
        (1, 0): [  # c2 hexagons mode 0;  final 2 pts in opp. mode.
            # If exterior, they are idx 2,1 (eg (0,0) (2,0)).
            [(3, 1), (2, 0), (0, 0), (-1, 1), (0, 2), (2, 2)],        # (0, 2), (2, 2)
            [(0, -2), (-1, -1), (0, 0), (2, 0), (3, -1), (2, -2)],    # (3, -1), (2, -2)
            [(-3, 1), (-1, 1), (0, 0), (-1, -1), (-3, -1), (-4, 0)]   # (-3, -1), (-4, 0)
        ],
        (1, 1): [  # c2 hexagons mode 1;  final 2 pts in opp. mode - exterior.
            [(-1, -1), (0, 0), (2, 0), (3, -1), (2, -2), (0, -2)],    # (2, -2), (0, -2)
            [(-1, 1), (0, 0), (-1, -1), (-3, -1), (-4, 0), (-3, 1)],  # (-4, 0), (-3, 1)
            [(2, 0), (0, 0), (-1, 1), (0, 2), (2, 2), (3, 1)]         # (2, 2), (3, 1)
        ],
        (2, 0): [  # region triangles mode 0
            [  # 0x26, 0x2a, 0x2b: c0 VΛV
                [(0, 0), (-1, 1), (1, 1)],  # 26
                [(1, 1), (2, 0), (0, 0)],  # 2a
                [(2, 0), (1, 1), (3, 1)],  # 2b
            ], [  # 0x3a, 0x39, 0x49:   c1 VΛV
                [(1, -1), (0, 0), (2, 0)],  # 3a
                [(0, 0), (1, -1), (-1, -1)],  # 39
                [(0, -2), (-1, -1), (1, -1)],  # 49
            ], [  # 0x35, 0x25, 0x21:   c2 VΛV
                [(-1, -1), (-2, 0), (0, 0)],  # 35
                [(-1, 1), (0, 0), (-2, 0)],  # 25
                [(-2, 0), (-3, 1), (-1, 1)],  # 21
            ]
        ],
        (2, 1): [  # region triangles mode 1
            [  # 0x39, 0x3a, 0x3e: c0  ΛVΛ
                [(0, 0), (1, -1), (-1, -1)],  # 39
                [(1, -1), (0, 0), (2, 0)],  # 3a
                [(2, 0), (3, -1), (1, -1)],  # 3e
            ], [  # 0x25, 0x35, 0x34: c1 ΛVΛ
                [(-1, 1), (0, 0), (-2, 0)],  # 25
                [(-1, -1), (-2, 0), (0, 0)],  # 35
                [(-2, 0), (-1, -1), (-3, -1)],  # 34
            ], [  # 0x2a, 0x26, 0x16: c2 ΛVΛ
                [(1, 1), (2, 0), (0, 0)],  # 2a
                [(0, 0), (-1, 1), (1, 1)],  # 26
                [(0, 2), (1, 1), (-1, 1)],  # 16
            ]
        ],
        (3, 0): [  # clockwise, c0,c1,c2 - edges of super-region
            [
                (-3, 1), (-1, 1), (1, 1),
                (3, 1), (2, 0), (1, -1),
                (0, -2), (-1, -1), (-2, 0),
            ]
        ],
        (3, 1): [  # clockwise, c0,c1,c2 - edges of super-region
            [
                (3, -1), (1, 1), (-1, -1),
                (-3, -1), (-2, 0), (-1, 1),
                (0, 2), (1, 1), (2, 0),
            ]
        ],
        (4, 0): [  # clockwise, c0,c1,c2 - vertexes of super-region
            [  # =  <-c0->  <-c1->   <-c2->
                (-3, 1), (3, 1), (0, -2),  # (-3, 1)
            ]
        ],
        (4, 1): [  # clockwise, c0,c1,c2 - vertexes of super-region
            [  # =  <-c0->    <-c1->  <-c2->
                (3, -1), (-3, -1), (0, 2)
            ]
        ],
        # Unshared points of cell excluding (0,0) use only on one mode
        # The other mode will be just the (0,0)
        (5, 0): [[(2, 0), (1, -1), (-1, -1), (-2, 0), (-1, 1), (1, 1)]]
    }

    if h9k is None:
        from hhg9.h9.constants import H9K
        h9k = H9K
    uv = np.array([h9k.lattice.U, 3 * h9k.lattice.V])
    hh = np.zeros((2, 3, 4, 2), dtype=np.float64)
    hx = np.zeros((2, 3, 6, 2), dtype=np.float64)
    tx = np.zeros((2, 3, 3, 3, 2), dtype=np.float64)
    te = np.zeros((2, 9, 2), dtype=np.float64)
    sr = np.zeros((2, 3, 2), dtype=np.float64)
    gd = np.zeros((6, 2), dtype=np.float64)
    for (kind, mode), c2s in pts.items():
        for c2, poly in enumerate(c2s):
            arr = np.asarray(poly, dtype=np.float64) * uv
            match kind:
                case 0:
                    hh[mode, c2] = arr
                case 1:
                    hx[mode, c2] = arr
                case 2:
                    tx[mode, c2] = arr
                case 3:
                    te[mode] = arr
                case 4:
                    sr[mode] = arr
                case 5:
                    gd = arr
    return H9Polygon(hh=hh, hx=hx, tx=tx, se=te, sv=sr, gd=gd)


H9P = _h9_polygon()


def region_grid(levels: int = 3, mode: int = 0, h9p: H9Polygon = H9P) -> List[Tuple]:
    """
    Generate a grid of points recursively.

    Returns:
        List of [address_path, current_mode, origin_xy, scale].
    """
    from hhg9.h9 import H9C, H9R
    h9c, h9r = H9C, H9R
    modes = [h9r.downs, h9r.ups]
    kids = modes[mode]
    queue = [(k, h9c.mode[k], h9c.off_xy[k], 1.0 / 3.0) for k in kids]

    for depth in range(levels):
        next_q = []
        for path, mode, origin, scale in queue:
            kids = modes[mode]  # shape (9,) indices
            for k in kids:
                mo_k = h9c.mode[k]
                off_k = h9c.off_xy[k]
                path_k = np.append(path, k)
                origin_k = origin + off_k * scale
                scale_k = scale / 3.0
                next_q.append((path_k, mo_k, origin_k, scale_k))
        queue = next_q
    return queue


def tri_grid(levels: int = 5, mode: int = 0, h9p: H9Polygon = H9P) -> NDArray[np.float64]:
    """
    Generate all triangle centroids for an octant at a given depth.

    Returns:
        NDArray: Shape (N, 3, 2) array of triangle vertices.
    """
    queue = region_grid(levels, mode, h9p)
    pts = np.empty((len(queue), 3, 2), dtype=np.float64)
    for i, (path, mode, origin, scale) in enumerate(queue):
        pts[i] = origin + h9p.sv[mode] * scale
    return pts


def tri_mesh(levels: int = 5, mode: int = 0, h9p: H9Polygon = H9P):
    """
    Return unique vertices and edges for the triangular mesh at a given level.

    Useful for creating Matplotlib Triangulations.

    Args:
        levels: Subdivision depth.
        mode: Root mode (0 or 1).
        h9p: Polygon LUT.

    Returns:
        tuple: (verts, edges, tris)
            verts: (V, 2) Unique vertex coordinates.
            edges: (E, 2) Edge indices into verts.
            tris: (T, 3) Triangle indices into verts.
    """
    # Use existing tri_grid to get all triangle vertices at this level
    tris_xy = tri_grid(levels=levels, mode=mode, h9p=h9p)  # (T, 3, 2)
    num_tris = tris_xy.shape[0]

    # Flatten all triangle vertices and deduplicate
    flat = tris_xy.reshape(-1, 2)  # (T*3, 2)
    verts, inv = np.unique(flat, axis=0, return_inverse=True)

    # Map each triangle to indices into the unique vertex array
    tris = inv.reshape(num_tris, 3)

    # Build undirected edge list from triangle connectivity
    e01 = tris[:, [0, 1]]
    e12 = tris[:, [1, 2]]
    e20 = tris[:, [2, 0]]
    edges_all = np.concatenate([e01, e12, e20], axis=0)

    # Sort each edge's endpoints so that (i, j) and (j, i) dedupe
    edges_sorted = np.sort(edges_all, axis=1)
    edges = np.unique(edges_sorted, axis=0)
    return verts, edges, tris


def enmesh(pts, levels: int = 35, shape=None):
    """
    Builds the half-hex polygons for a batch of points up to `levels` depth.

    Walks the hierarchy for each point and generates the geometry for every layer.

    Returns:
        tuple: (uniques, refs)
            uniques: (U, 4, 2) Unique polygons in global coordinates.
            refs: (N, depth) Indices into `uniques` for each input point/layer.
    """
    from hhg9.h9 import H9C
    from hhg9.h9.region import H9R, xy_regions_iter

    # Modes per point
    oc, mo = pts.cm()
    coords = pts.coords
    num_pts = len(pts)
    depth = levels + 1

    # Offsets and shapes (float64)
    offs = H9C.off_xy.astype(np.float64, copy=False)  # (R,2)
    hh = H9P.hh  # (2,3,4,2)

    # Accumulators
    parent_xy = np.zeros((num_pts, 2), dtype=np.float64)
    scale = 1.0

    # We'll collect per-layer polygons flattened as (N*D, 4, 2)
    polys = np.empty((num_pts * depth, 4, 2), dtype=np.float64)
    flat_idx = np.empty((num_pts, depth), dtype=np.int64)  # indices into `polys`

    for ev in xy_regions_iter(coords, mode=mo, depth=levels):
        if ev.phase != 'pre':
            continue
        i = ev.i  # 0..D-1

        # Child c2 for each row under the parent mode
        c2 = H9R.mcc2[ev.pmo, ev.cid]  # (N,)
        hh_shape = hh[ev.pmo, c2]  # pmo

        # Translate/scale each triangle to global coords
        polygon = parent_xy[:, None, :] + hh_shape * scale  # (N,4,2)

        # Store flattened by layer, keeping a stable mapping back to rows
        start = i * num_pts
        polys[start:start + num_pts] = polygon
        flat_idx[:, i] = np.arange(start, start + num_pts, dtype=np.int64)

        # Step parent origin for next layer
        parent_xy = parent_xy + offs[ev.cid] * scale
        scale /= 3.0

    if num_pts > 1:
        # Deduplicate exactly: reshape to (M, 8) and unique over rows
        mass = num_pts * depth
        flat = polys.reshape(mass, 8)
        uniq, first_idx, inv = np.unique(flat, axis=0, return_index=True, return_inverse=True)
        uniques = uniq.reshape(-1, 4, 2)

        # Map each (address, layer) to its unique polygon index
        refs = inv[flat_idx]
        return uniques, refs
    else:
        return polys, np.array(list[range(num_pts)])


class HexReducer(Protocol):
    """Callable signature for aggregating per-point data into per-hex values."""

    def __call__(
        self,
        inv_hex: NDArray[np.int64],
        counts: NDArray[np.int64],
        pts,
    ) -> NDArray[np.float64]:
        ...


def hex_poly_groups(pts, layers: int = 10, tail_style: TailStyle = TailStyle.reversible):
    """Return hex polygons plus the grouping needed to aggregate arbitrary per-point data.

    Returns:
        tuple: (hx_pts, inv_hex, counts, idx)
            hx_pts: Points of hex polygon vertices (H*6,2) with per-vertex octant components.
            inv_hex: (N,) mapping each input point -> hex index in [0, H)
            counts: (H,) population count per hex
            idx: (H,) indices of representative points for each hex (as returned by np.unique)

    Notes:
        - This function does NOT aggregate values; callers can compute means/sums/modes/medians etc.
        - `idx` can be used to recover per-hex address metadata from the representative point.
    """
    from hhg9 import Points

    b_oct = pts.domain
    if b_oct.name != 'b_oct':
        raise ValueError('hex_poly_groups requires pts to be in b_oct domain')

    h_val = hex_layer(pts, layer=layers, tail_style=tail_style)
    h_key = hex_key(h_val)

    hex_k, idx, inv_hex = np.unique(h_key, axis=0, return_index=True, return_inverse=True)
    hex_num = hex_k.shape[0]
    counts = np.bincount(inv_hex, minlength=hex_num).astype(np.int64, copy=False)

    # Build hex polygons (same logic as the old hex_poly_layer)
    hex_v = h_val[idx]
    tail = hex_v[:, -1]
    xpm, xc2, xrm, rgn = tail_unpack_reversible(tail)
    hex_all = H9P.hx[xpm, xc2]
    hex_xy = np.zeros((hex_num, 2), dtype=float)
    hex_oid, hex_rgn = hex_digits_reg(hex_v, b_oct)

    scale = 1.0
    for i in range(1, layers + 1):
        hex_xy += H9C.off_xy[hex_rgn[:, i]] * scale
        scale /= 3.0

    hex_pts = hex_xy[:, None, :] + hex_all * scale  # (H,6,2)

    # Set the basic octant id for each hexagon
    oc_poly6 = np.repeat(hex_oid[:, None], 6, axis=1)  # (H, 6)
    nbr_oid = b_oct.oid_nb[hex_oid, xc2]

    hex_ẋ = H9K.R3 * hex_pts[..., 0].ravel()  # Classifier ẋ
    hex_y = hex_pts[..., 1].ravel()          # Classifier y
    oc_mo = np.repeat(xrm, 6)                # Set the octant mode for each hexagon
    types = location(hex_ẋ, hex_y, oc_mo)
    locs = types.reshape(-1, 6)

    ex4 = locs[:, 4] == 1
    ex5 = locs[:, 5] == 1
    ext_hex = (ex4 & ex5)

    # Stitch exterior vertices across face boundaries
    hex_pts[ext_hex, 4] = hex_pts[ext_hex, 2] * [1, -1]
    hex_pts[ext_hex, 5] = hex_pts[ext_hex, 1] * [1, -1]

    n_oct = nbr_oid[ext_hex]
    oc_poly6[ext_hex, 4] = n_oct
    oc_poly6[ext_hex, 5] = n_oct

    hx_coords = hex_pts.reshape([-1, 2])
    hx_oc = oc_poly6.reshape([-1])
    hx_pts = Points(hx_coords, b_oct, components=hx_oc)

    return hx_pts, inv_hex.astype(np.int64, copy=False), counts, idx



def hex_poly_layer(pts, layers: int = 10, reducer: Optional[HexReducer] = None):
    """Return hex polygons and an aggregated value per hex.

    By default, this reproduces the previous behaviour: mean of `pts.samples` per hex.

    Args:
        pts: Input points in b_oct.
        layers: Hex layer.
        reducer: Optional callable to aggregate per-point data into per-hex values.
            Signature: reducer(inv_hex, counts, pts) -> (H,) float array.

    Returns:
        tuple: (hx_pts, values)
            hx_pts: Points of hex polygon vertices (H*6,2)
            values: (H,) aggregated value per hex (float64)
    """
    hx_pts, inv_hex, counts, idx = hex_poly_groups(pts, layers=layers, tail_style=TailStyle.reversible)

    if reducer is None:
        # Default: mean of pts.samples per hex (previous behaviour)
        sum_wt = np.bincount(inv_hex, weights=pts.samples, minlength=counts.shape[0])
        values = np.divide(sum_wt, counts, out=np.zeros_like(sum_wt, dtype=float), where=counts > 0)
    else:
        values = reducer(inv_hex, counts, pts)
        values = np.asarray(values, dtype=np.float64)
        if values.shape != (counts.shape[0],):
            raise ValueError(f"reducer must return shape ({counts.shape[0]},), got {values.shape}")

    return hx_pts, values


def hh_layer(pts, layer: int = 10):
    """
    Calculates Half-Hexagon polygons for a batch of points.

    Returns:
        tuple: (polygon, pops, inv, oc_poly4)
            polygon: (H, 4, 2) Half-hex vertices.
            pops: Population count per half-hex.
            inv: Inverse mapping.
            oc_poly4: Octant IDs for the 4 vertices.
    """
    from hhg9.h9.region import xy_regions
    from hhg9.h9 import H9C, H9R
    num_pts = len(pts)
    if num_pts == 0:
        return

    ocf, mode_f = pts.cm()  # return the octant/mode
    rgx = xy_regions(pts.coords, mode_f, layer)
    poi = rgx[:, -2]
    smo = H9C.mode[poi]  # We will need it's mode (smo=self_mode)
    c2i = rgx[:, -1]
    c2 = H9R.mcc2[smo, c2i]  # Now grab the c2 of the current region - which neighbour?!
    key = np.concatenate([ocf[:, None], rgx[:, 1:layer + 1], c2[:, None]], axis=1)
    uniq, first_idx, inv = np.unique(key, axis=0, return_index=True, return_inverse=True)
    pops = np.bincount(inv)
    rgs = rgx[first_idx]  # This gives us the indicative regions.
    oc = ocf[first_idx]  # and the indicative octants for each region.
    c2 = c2[first_idx]
    num_hh = rgs.shape[0]  # and the number of half-hexagons found.
    acc_xy = np.zeros((num_hh, 2), dtype=np.float64)
    scale = 1.0
    for i in range(1, layer + 1):
        acc_xy += (H9C.off_xy[rgs[:, i]] * scale)
        scale /= 3.0
    poi = rgs[:, -2]
    smo = H9C.mode[poi]  # We will need its mode (smo=self_mode)
    polygon = scale * H9P.hh[smo, c2] + acc_xy[:, None]
    oc_poly4 = np.repeat(oc[:, None], 4, axis=1)  # 4 points of half-hex
    return polygon, pops, inv, oc_poly4


# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
A Domain is a numerical coordinate space used inside Hex9.

It defines:
  • The coordinate axes (their order, units, and interpretation),
  • The region of that space that is considered valid (via `valid`),
  • The default point / address formats for representing those coordinates.

A Domain is **not** a CRS: it has no datum, no ellipsoid choice, and no external
projection baked in. Instead, Domains are stitched together by separate
Projection objects (e.g. g_gcd ↔ c_oct ↔ b_oct), which map between Domains
and thus between Hex9 spaces and real-world geodetic coordinates.

If we have some latitude/longitude coordinates in a numpy array, we use the 'g_gcd' Domain from Registrar,
and then use that to instantiate the coordinates into a Projection-ready Points classs.

reg=Registrar()
gcd_pts=Points(my_coords, reg.domain('g_gcd'))    # <-- domain tells Points what my_coords represent.
ell_pts=reg.project(gcd_pts, ['c_ell'])           # can now project lat-lon to ECEF xyz.

ellipsoidal surface of the WGS84 ellipsoid using a GCD projection.
"""
from abc import ABC, abstractmethod
import numpy as np
from numpy.typing import NDArray

from .point_format import PointFormat


class Domain(ABC):
    """
    Base class for a Domain.
    This is not a CRS, because it does not include projections.
    """

    def __init__(self, registrar, name: str, axes):
        self.name = name
        self.axes = axes
        if registrar is not None:
            registrar.register_domain(self)
        self.default_format = None
        self.address_formats = {}  # class static

    def sig(self) -> str:
        """
        :return: my name.
        """
        return self.name

    def register_format(self, af: PointFormat):
        """Decorator to register an AddressFormat"""
        self.address_formats[af.name] = af
        af._set = self

    @abstractmethod
    def valid(self, _) -> NDArray:
        """Return a set of bools for those points which are valid in this set."""
        ...

    def is_valid(self, pts) -> bool:
        """
        :param pts: set of points for in this set.
        :return: that the points are all valid.
        """
        return np.all(self.valid(pts))

    def where_valid(self, pts):
        """
        :param pts: set of points for in this set.
        :return: those which are legal.
        """

        vx = self.valid(pts)
        return np.array(pts)[vx]

    def adopt(self, pts: NDArray):
        """
        Take an array and adopt as this domain.
        """
        from .points import Points
        good = self.where_valid(pts)
        return Points(good, self)

    def __repr__(self):
        return f'{self.name}'




# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
H9 Constants (Single Source of Truth).

This module declares H9 constants **once** in float64 and freezes them.
This acts as the single source of truth for radicals like :math:`\\sqrt{3}`, :math:`\\sqrt{2}`,
and derived geometric bounds.

**Why this exists:**
    Calculated values (like :math:`\\sqrt{3} / 2`) can vary slightly depending on the order of
    operations due to float64 ULP (Unit in Last Place) variations. By defining them strictly
    here, we ensure stability across downstream edge cases and inequality tests.

**Notation:**
    * **Λ (Lambda):** Denotes Supercell Mode 1 ("Up" or "Peak").
    * **V:** Denotes Supercell Mode 0 ("Down" or "Valley").
    * **C/F:** Suffixes for Ceiling/Floor in barycentric orientation (not Apex/Base!).
    * **Supercell Barycentre:** Fixed at 0; each supercell spans zero.

See ``constants.md`` for further documentation.
"""

from dataclasses import dataclass
import numpy as np


@dataclass(frozen=True, slots=True)
class Fundamentals:
    """
    Base radical constants.

    Because of variations found in float64 ULP (units of least precision),
    we calculate the constants just once.

    Attributes:
        R3 (float): The square root of 3 (:math:`\\sqrt{3}`).
        W (float): The square root of 2 (:math:`\\sqrt{2}`), representing the
            edge length of the supercell triangle.
    """
    R3: float
    W: float


@dataclass(frozen=True, slots=True)
class Derived:
    """
    Values derived once from Fundamentals.

    These are used across inequalities and coordinate transforms.

    Attributes:
        W (float): Edge length of the *full* barycentric triangle (:math:`\\sqrt{2}`).
            Mirrors ``Fundamentals.W`` for convenience.
        H (float): Height of the supercell triangle. Calculated as :math:`H = W \\frac{\\sqrt{3}}{2}`.
        Ḣ (float): Height of a single child cell (1/3 of total Height). :math:`\\dot{H} = H / 3`.
        Ẇ (float): Width parameter for child cells, used in inequality tests. :math:`\\dot{W} = 2 \\dot{H}`.
            Also acts as an alias for ``ΛC``.
        RH (float): Ratio of Height to Width (:math:`\\frac{\\sqrt{3}}{2}`). Useful for symmetric transforms.
    """
    W: float
    H: float
    Ḣ: float
    Ẇ: float
    RH: float


@dataclass(frozen=True, slots=True)
class Limits:
    """
    Horizontal (TR, TL) and Vertical bounds for Λ (Mode 1) and V (Mode 0) triangles.

    These values define the bounding box of the triangles in the classifier plane.
    Vertical values are defined relative to the barycenter (y=0).

    Attributes:
        TR (float): Positive right-hand x-limit in classifier plane (:math:`+W/2`).
        TL (float): Negative left-hand x-limit in classifier plane (:math:`-W/2`).
        ΛC (float): Mode 1 (Up) supercell barycentric ceiling. :math:`y = +2\\dot{H}`.
        VC (float): Mode 0 (Down) supercell barycentric ceiling. :math:`y = +1\\dot{H}`.
        ΛF (float): Mode 1 (Up) supercell barycentric floor. :math:`y = -1\\dot{H}`.
        VF (float): Mode 0 (Down) supercell barycentric floor. :math:`y = -2\\dot{H}`.
    """
    TR: float
    TL: float
    ΛC: float
    VC: float
    ΛF: float
    VF: float


@dataclass(frozen=True, slots=True)
class Lattice:
    """
    Rectilinear lattice steps.

    Uses multiples of (U, V) as unit multipliers to calculate cell offsets.
    The barycentre of every equilateral triangle (cell) tiling the plane may be
    identified using integer co-ordinates (x, y) such that its Cartesian position
    is at :math:`(xU, yV)`.

    Note:
        Not every (u, v) combination will be a valid barycentre, but every
        valid barycentre is an integer coordinate.

    Attributes:
        U (float): Horizontal unit step (:math:`W / 6`).
        Ü (float): Scaled horizontal unit (:math:`\\sqrt{3} U`). Note: Effectively equal to ``Ḣ``.
        V (float): Vertical unit step (:math:`H / 9`).
    """
    U: float
    Ü: float
    V: float


@dataclass(frozen=True, slots=True)
class H9Const:
    """
    The master container for H9 Grid Constants.

    This class aggregates Fundamentals, Derived values, Limits, and Lattice steps.
    It provides property accessors for the most commonly used values to simplify
    import statements in other modules.
    """
    radical: Fundamentals
    derived: Derived
    limits: Limits
    lattice: Lattice

    @property
    def H(self) -> float:
        """Height of the supercell triangle."""
        return self.derived.H

    @property
    def R3(self) -> float:
        """Square root of 3."""
        return self.radical.R3

    @property
    def ΛF(self) -> float:
        """Mode 1 (Up) floor."""
        return self.limits.ΛF

    @property
    def ΛC(self) -> float:
        """Mode 1 (Up) ceiling."""
        return self.limits.ΛC

    @property
    def VF(self) -> float:
        """Mode 0 (Down) floor."""
        return self.limits.VF

    @property
    def VC(self) -> float:
        """Mode 0 (Down) ceiling."""
        return self.limits.VC

    @property
    def U(self) -> float:
        """Lattice horizontal unit."""
        return self.lattice.U

    @property
    def V(self) -> float:
        """Lattice vertical unit."""
        return self.lattice.V

    @property
    def Ü(self) -> float:
        """Lattice scaled horizontal unit (Unicode U with diaeresis)."""
        return self.lattice.Ü

    @property
    def Ẇ(self) -> float:
        """Child cell width parameter (Unicode W with dot above)."""
        return self.derived.Ẇ

    @property
    def Ḣ(self) -> float:
        """Child cell height parameter (Unicode H with dot above)."""
        return self.derived.Ḣ


def h9_constants() -> H9Const:
    """
    Factory method to initialize the H9Const singleton.

    Calculates all derived values from the two fundamental roots to ensure
    floating-point consistency.

    Returns:
        H9Const: The frozen, initialized constants object.
    """
    # Define the Fundamental constants
    # In this grid, the fundamental edge-length of the supercell triangle is √2
    # due to its relation with the unit-octahedral shape of the underlying system.
    # We choose W (√2) as a fundamental alongside R3 (√3).
    r3 = np.sqrt(3.0, dtype=np.float64)  # √3 because equilateral triangles.
    w = np.sqrt(2.0, dtype=np.float64)   # √2 from unit-octahedron edge length
    k_fundamentals = Fundamentals(r3, w)

    # Define the Derived constants
    h = w * r3 / 2.0  # supercell triangle height via Pythagoras
    ḣ = h / 3.        # Vertical spacing between stacked child cells: Ḣ = H / 3
    ẇ = 2 * ḣ         # slope-window used in y±ẋ tests; alias of ΛC, ẇ is 2/3 of h
    rh = r3 / 2.      # ratio of height to width. 0.8660
    k_derived = Derived(w, h, ḣ, ẇ, rh)

    # Define the Limits constants
    tr = h / r3
    tl = -tr
    # Limits declares ceilings / floors in barycentric order descending.
    λc = +ẇ   # mode 1 (up) triangle ceiling   y: +2·Ḣ
    vc = +ḣ   # mode 0 (down) triangle ceiling y: +1·Ḣ
    λf = -ḣ   # mode 1 (up) triangle floor     y: -1·Ḣ
    vf = -ẇ   # mode 0 (down) triangle floor   y: -2·Ḣ
    k_limits = Limits(tr, tl, λc, vc, λf, vf)

    # Define the Lattice constants
    # Lattice (small grid steps; parent step = (3U, 3V))
    u, v = w / 6., h / 9.  # horizontal/vertical unit multipliers.
    # Note: The second argument passed to Lattice here is ḣ (which equals √3 * u)
    k_lattice = Lattice(u, ḣ, v)

    return H9Const(
        radical=k_fundamentals,
        derived=k_derived,
        limits=k_limits,
        lattice=k_lattice,
    )


H9K: H9Const = h9_constants()  # singleton for typical use

# Tolerant checks (float64-safe)
assert np.isclose(H9K.ΛF, -H9K.VC, rtol=1e-12, atol=1e-15)
assert np.isclose(H9K.ΛC, -H9K.VF, rtol=1e-12, atol=1e-15)

# Ordering sanity
assert H9K.ΛC > 0.0 > H9K.ΛF
assert H9K.VC > 0.0 > H9K.VF

# Span consistency (optional)
# total span in Λ equals total span in V by construction:
assert np.isclose((H9K.ΛC - H9K.ΛF), (H9K.VC - H9K.VF), rtol=1e-12, atol=1e-15)

# Apex farther than base; combined band ordering
assert H9K.ΛC > H9K.VC > 0.0 > H9K.ΛF > H9K.VF



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Useful Geometry functions
"""
import numpy as np


def inside_triangle_cw(pts: np.ndarray, tri: np.ndarray) -> np.ndarray:
    """Vectorised point-in-triangle test.
    points: (N,2), tri: (3,2) in CW order. Returns mask bool."""
    # Edges
    a, b, c = tri[0], tri[1], tri[2]
    ab = b - a
    bc = c - b
    ca = a - c
    ap = pts - a
    bp = pts - b
    cp = pts - c
    cross1 = ab[0] * ap[:, 1] - ab[1] * ap[:, 0]
    cross2 = bc[0] * bp[:, 1] - bc[1] * bp[:, 0]
    cross3 = ca[0] * cp[:, 1] - ca[1] * cp[:, 0]
    return (cross1 <= 0) & (cross2 <= 0) & (cross3 <= 0)


def ortho_basis_from_normal(n):
    """Return two unit vectors (u, v) that form an orthonormal basis of the
    tangent plane given a unit normal vector n (shape (3,)).
    """
    nx, ny, nz = n
    # pick a vector not parallel to n
    if abs(nx) < 0.5:
        a = np.array([1.0, 0.0, 0.0])
    else:
        a = np.array([0.0, 1.0, 0.0])
    u = np.cross(n, a)
    u /= np.linalg.norm(u)
    v = np.cross(n, u)
    v /= np.linalg.norm(v)
    return u, v


def ellipsoid_f_grad(p, a, b):
    """Implicit surface function f(x,y,z)=x^2/a^2 + y^2/a^2 + z^2/b^2 - 1
    and its gradient at p (shape (3,)).
    Returns (f, gradF).
    """
    x, y, z = p
    inv_a2 = 1.0 / (a * a)
    inv_b2 = 1.0 / (b * b)
    f = x * x * inv_a2 + y * y * inv_a2 + z * z * inv_b2 - 1.0
    grad = np.array([2.0 * x * inv_a2, 2.0 * y * inv_a2, 2.0 * z * inv_b2])
    return f, grad



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Projection is abstract base class for managing Points conversions.
For general use, this is not publicly referred to - as projections are normally managed by the Registrar.
"""
from abc import ABC
from numpy.typing import NDArray
from .points import Points


class Projection(ABC):
    """
    Abstract class that manages A<=>B coordinate set conversion.
    This will typically include two methods - forward and backward,
    which are inverses of each other but do not necessarily indicate
    alliance to one side or the other
    """

    def __init__(self, registrar, name: str, fwd, rev):
        self.name = name
        self.fwd_cs = registrar.domain(rev)
        self.rev_cs = registrar.domain(fwd)
        registrar.register_projection(self, [fwd, rev])

    def forward(self, pts: Points) -> NDArray:
        """Move from one coordinate set to another; the inverse of backward."""
        ...

    def backward(self, pts: Points) -> NDArray:
        """Move from one coordinate set to another; the inverse of forward."""
        ...



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

import numpy as np
from numpy.typing import NDArray
from hhg9 import Points
from hhg9.base.projection import Projection
from pyproj import CRS, Transformer


class EllipsoidGCD(Projection):
    """
    Ellipsoidal XYZ <=> LatLon projection using WGS84.
    Forward: ECEF → (lat, lon)
    Backward: (lat, lon) → ECEF
    """

    def __init__(self, registrar):
        super().__init__(registrar, 'ell_gcd', 'c_ell', 'g_gcd')
        # Define coordinate systems
        c_ell = registrar.domain('c_ell')
        crs_geodetic = CRS.from_epsg(4326)  # WGS84 lat/lon
        # pyproj Transformers
        self.to_geodetic = Transformer.from_crs(c_ell.crs, crs_geodetic, always_xy=True)
        self.to_ecef = Transformer.from_crs(crs_geodetic, c_ell.crs, always_xy=True)

    def forward(self, arr: Points) -> NDArray:
        """ECEF → Geodetic (lat, lon)"""
        x, y, z = arr.coords[..., 0], arr.coords[..., 1], arr.coords[..., 2]
        lon, lat, _ = self.to_geodetic.transform(x, y, z)
        latlon = np.stack([lat, lon], axis=-1)
        return Points(latlon, domain=self.fwd_cs, samples=arr.samples)

    def backward(self, arr: Points) -> NDArray:
        """Geodetic (lat, lon) → ECEF"""
        lat, lon = arr.coords[..., 0], arr.coords[..., 1]
        heights = np.zeros_like(lat)
        x, y, z = self.to_ecef.transform(lon, lat, heights)
        xyz = np.stack([x, y, z], axis=-1)
        return Points(xyz, domain=self.rev_cs, samples=arr.samples)


# Use EllipsoidGCDRad when you want g_gcd in radians (r_gcd) to avoid per-candidate deg↔rad churn in root-finding.
# The degree-based EllipsoidGCD remains for human-facing I/O and compatibility.
class EllipsoidGCDRad(Projection):
    """
    Ellipsoidal XYZ ↔ Geodetic (lat, lon in radians) using WGS84.
    Forward: ECEF → (lat_rad, lon_rad)
    Backward: (lat_rad, lon_rad) → ECEF
    """
    def __init__(self, registrar):
        super().__init__(registrar, 'ell_gcr', 'c_ell', 'r_gcd')
        # Ensure r_gcd domain and deg↔rad projections exist (lazy/idempotent)
        try:
            registrar.domain('r_gcd')
        except Exception:
            from hhg9.domains import RadiansGCD
            from hhg9.projections import RGCD_GCD
            RadiansGCD(registrar)
            RGCD_GCD(registrar)
        # Reuse the pyproj transformers from the degree-based class
        crs_geodetic = CRS.from_epsg(4326)
        crs_ecef = CRS.from_epsg(4978)
        self.to_geodetic = Transformer.from_crs(crs_ecef, crs_geodetic, always_xy=True)
        self.to_ecef = Transformer.from_crs(crs_geodetic, crs_ecef, always_xy=True)

    def forward(self, arr: Points) -> NDArray:
        """ECEF → Geodetic radians (lat, lon in rad)"""
        x, y, z = arr.coords[..., 0], arr.coords[..., 1], arr.coords[..., 2]
        lon_deg, lat_deg, _ = self.to_geodetic.transform(x, y, z)
        latlon_rad = np.radians(np.stack([lat_deg, lon_deg], axis=-1))
        return Points(latlon_rad, domain=self.fwd_cs, samples=arr.samples)

    def backward(self, arr: Points) -> NDArray:
        """Geodetic radians (lat, lon in rad) → ECEF"""
        lat_deg = np.degrees(arr.coords[..., 0])
        lon_deg = np.degrees(arr.coords[..., 1])
        heights = np.zeros_like(lat_deg)
        x, y, z = self.to_ecef.transform(lon_deg, lat_deg, heights)
        xyz = np.stack([x, y, z], axis=-1)
        return Points(xyz, domain=self.rev_cs, samples=arr.samples)


if __name__ == '__main__':
    from support import Display, Photo
    from hhg9 import Registrar
    from hhg9.domains import GeneralGCD, PlatePixel, EllipsoidCartesian
    from hhg9.projections import PlatePixelGCD

    reg = Registrar()
    p_pix = PlatePixel(reg)
    e_sph = EllipsoidCartesian(reg)
    g_gen = GeneralGCD(reg)      # (xyz)
    ppg = PlatePixelGCD(reg)
    prj = EllipsoidGCD(reg)   # EllipsoidalGCD<->Cartesian (xyz)

    ps = Photo()

    ps.load('../../preparatory/world1350x675.png')
    pc_px = p_pix.adopt(ps.img)
    gg_ll = reg.project(pc_px, [p_pix, g_gen])
    gg_el = reg.project(gg_ll, [g_gen, e_sph])

    d = Display()  # simple support display class
    d.show_pts_3d(gg_el)



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Part of the H9 project
"""
import numpy as np
from numpy.typing import NDArray
from hhg9 import Points
from hhg9.base.projection import Projection


class PlatePixelGCD(Projection):
    """
        Convert Plate Carrée pixel to latitude/longitude
    """

    def __init__(self, registrar):
        super().__init__(registrar, 'pix_gcd', 'p_pix', 'g_gcd')
        self.e = 1e-100
        self.p_hgt = None
        self.p_wid = None
        self.lon = None
        self.lat = None

    def set_dim(self, pts: Points = None, bounds: object = None) -> object:
        """Set dimensions for plate."""
        if self.p_hgt is None or pts is not None:
            px, py = pts.coords[:, 0], pts.coords[:, 1]
            self.p_hgt = np.uint32(np.max(py)+1)
            self.p_wid = np.uint32(np.max(px)+1)
        if bounds is None:
            self.lon = np.linspace(-180+self.e, 180-self.e, self.p_wid)
            self.lat = np.linspace(-90, 90, self.p_hgt)
        else:
            # Bounds: (lon_min, lon_max, lat_min, lat_max)
            lon_min, lon_max, lat_min, lat_max = bounds
            self.lon = np.linspace(lon_min+self.e, lon_max-self.e, self.p_wid)
            self.lat = np.linspace(lat_min, lat_max, self.p_hgt)

    def forward(self, pts: Points) -> NDArray:
        """
        INPUT:  Plate Carrée coordinates (origin bottom-left)
        OUTPUT: GCD
        """
        if self.lon is None:
            self.set_dim(pts)
        px, py = pts.coords[:, 0].astype(np.uint32), pts.coords[:, 1].astype(np.uint32)
        lx = self.lon[px]  # longitude is <w--e>
        ly = self.lat[py]  # latitude is <n--s>
        ret = np.stack([ly, lx], dtype=np.float64, axis=-1)
        return Points(ret, domain=self.fwd_cs, samples=pts.samples)

    def backward(self, pts: Points) -> NDArray:
        """
        Convert (lat, lon) to pixel (x, y) coordinates using the
        lookup tables created by set_dim.
        """
        if self.lon is None:
            self.set_dim(pts)

        # Coordinates are in (lat, lon) order from GCD domain
        la = pts.coords[:, 0].astype(np.float64)
        lo = pts.coords[:, 1].astype(np.float64)

        # Use the pre-calculated lookup tables from set_dim.
        # Do NOT recalculate them here. These are correctly bounded to your map.
        px = np.searchsorted(self.lon, lo)
        py = np.searchsorted(self.lat, la)

        # Return pixel coordinates in (x, y) order
        ret = np.stack([px, py], axis=-1)
        return Points(ret, domain=self.rev_cs, samples=pts.samples)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    from hhg9 import Registrar
    from hhg9.domains import EllipsoidCartesian, GeneralGCD, PlatePixel

    reg = Registrar()
    p_pix = PlatePixel(reg)
    c_sph = EllipsoidCartesian(reg)  # Cartesian Spherical (xyz)
    g_sph = GeneralGCD(reg)  # Cartesian Spherical (xyz)
    pg = PlatePixelGCD(reg)

    # # Create dummy image of shape (18, 36, 3)
    h, w = 1800, 3600
    img = np.zeros((h, w, 1))
    p0 = p_pix.adopt(img)  # Shape: (648, 5)
    # p0 = np.array([[i, 1799] for i in range(3600)])

    # Project to lat/lon and back
    l1 = reg.project(p0, [p_pix, g_sph])
    p1 = reg.project(l1, [g_sph, p_pix])

    # Compute pixel round-trip error
    original_px = np.array(p0.coords, dtype=np.uint64)
    projects_px = np.array(p1.coords, dtype=np.uint64)

    px_error = np.linalg.norm(original_px - projects_px, axis=1)
    p1.samples = px_error
    p2 = p_pix.image(p1)
    plt.imshow(p2, origin='lower')
    plt.show()

    plt.imshow(p2, cmap='hot', origin='lower')
    plt.colorbar(label="Pixel error (L2 norm)")
    plt.title("Round-trip pixel error (Plate Carrée)")
    plt.show()

    error_img = px_error.reshape(h, w)
    plt.imshow(error_img, cmap='hot', origin='lower')
    plt.colorbar(label="Pixel error (L2 norm)")
    plt.title("Round-trip pixel error (Plate Carrée)")
    plt.show()



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Part of the H9 project
"""
import numpy as np
from hhg9 import Registrar
from hhg9.base.point_format import PointFormat
from hhg9.base.points import Points
import hhg9.h9.addressing as adr
from hhg9.algorithms.packing import u64_pack


class OctahedralH9(PointFormat):
    """
    Addressing of a half-hexagonal grid over OctahedralBarycentric.
    Handles the eight face names (NEA, …). Implements the PointFormat contract:
      - __init__(name)
      - format(points, composite, sub)
      - revert(address: str|list[str]) → Points
    """
    def __init__(self, registrar: Registrar) -> None:
        super().__init__(registrar, 'h9')
        self.scheme = adr.H9_RA
        self.width = 34  # default printed width
        self.subs = {
            'k': adr.TailStyle.key,
            'x': adr.Style.HEX,
            'i': adr.Style.NUMERIC,
            'u': adr.Style.UH64,
            'r': adr.Style.UR64,
        }

    def is_valid(self, address: str) -> bool:
        """Required by PointFormat; not implemented yet."""
        return True  # placeholder until a validator is added

    def revert(self, address, style=adr.Style.HEX):
        """
        :return: bary addresses(es) as Points.
        """
        result = None

        if isinstance(address, str):
            if '\n' in address:
                arr = address.splitlines()
            else:
                arr = [address]
            if style != adr.Style.HEX:
                dgt = [([int(a, 16) for a in adx]) for adx in arr]
                arr = u64_pack(np.array(dgt))
            else:
                arr = np.array(arr)
        match style:
            case adr.Style.HEX:
                result = adr.hex_str_decode(arr, self.registrar, self.scheme)
            case adr.Style.UH64:
                result = adr.hex_unpack(arr, self.registrar, self.scheme)
            case adr.Style.NUMERIC:
                result = adr.hex_unpack(arr, self.registrar, self.scheme)
            case adr.Style.UR64:
                result = adr.reg_unpack(arr, self.registrar, self.scheme)
        return result

    def _select_style(self, sub: str = None):
        """
        Given a string, determine format style
        eg h9.f33, h9.u18, h9.24
        """
        width = self.width
        style = adr.Style.HEX
        # Parse sub: optional style letter(s) + optional integer width
        if sub:
            # long token 'u' takes precedence; otherwise first char
            if sub.startswith('u'):
                width = 13
                style = self.subs['u']
                sub = sub[1:]
            elif sub.startswith('r'):
                width = 14
                style = self.subs['r']
                sub = sub[1:]
            elif sub[0] in self.subs:
                style = self.subs[sub[0]]
                sub = sub[1:]
            # numeric tail → width
            if sub:
                try:
                    width = int(sub)
                except Exception:
                    raise ValueError(f"h9f format: invalid width '{sub}' — expected integer (e.g. :h9f.33)")
        return style, width

    def format(self, arr: Points, _, sub: str):
        """
        Return H9 label(s) for the given Points.
        The `sub` string (from Points.__format__) may contain an optional style prefix
        and/or a decimal width, e.g. 'x33', 'i21', '33'. Default: hex body, width=self.width.
        """
        # if self.engine is None:
        #     self.engine = arr.domain.engine
        if not isinstance(arr, Points):
            pts = arr
        else:
            pts = arr
        tail_style = adr.TailStyle.reversible
        count = len(pts)
        style, width = self._select_style(sub)
        if style == adr.TailStyle.key:
            tail_style = adr.TailStyle.key

        if style == adr.Style.UH64:
            u64 = adr.hex_pack(pts, width, self.scheme)  # 3 metres - not great!
            strs = [''.join([f'{u:0x}' for u in v])[:width+4] for v in u64]
            if count < 2:
                return strs[0]
            return '\n'.join(strs)

        if style == adr.Style.UR64:
            u64 = adr.reg_pack(pts, width, self.registrar, self.scheme)  # 0.5 metres - ok!
            strs = [''.join([f'{u:016x}' for u in v])[:width+2] for v in u64]
            if count < 2:
                return strs[0]
            return '\n'.join(strs)

        if style == adr.Style.NUMERIC:
            u64 = adr.hex_pack(pts, width, self.scheme)
            strs = [''.join(f"{n:0x}" for n in row) for row in u64]
            if count < 2:
                return str(strs[0])
            return '\n'.join(n for n in strs)

        h9h = adr.hex_str_encode(pts, width, tail_style, self.scheme)
        if count < 2:
            return str(h9h[0])
        return '\n'.join(n for n in h9h)



"""
Part of the H9 project
e.g. Domain-specific, but widely reusable put it in domains/
Core infrastructure (used across points, formats, projections, etc.) → put it in base/
H9 Subpackage Initialization.
Order matters here to avoid circular imports.
"""

from .constants import H9Const, H9K
from .classifier import H9Classifier, H9CL, classify_cell, in_scope, in_down, in_up
from .lattice import H9C, H9Cell
from .region import H9R, H9Region
from .addressing import H9_RA, HEX_LUTS
from .polygon import H9P, H9Polygon

__all__ = [
    "H9K", "H9Const",
    "H9CL", "H9Classifier",
    "H9C", "H9Cell",
    "H9R", "H9Region",
    "H9P", "H9Polygon",
    "H9_RA", "HEX_LUTS",
    "classify_cell",
    "in_scope", "in_down", "in_up"
]



# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
Trace utilities for Hex9 (H9) project. Useful when debugging.
"""
import os
import traceback


def caller(msg):
    """Print the caller. Skip this, and the call to this!"""
    dx = traceback.extract_stack()[-3]
    sx = f'{msg}: {os.path.relpath(dx.filename, "..")}, line {dx.lineno}; at {dx.line}'
    print(sx)
